''' opencos.tools.vivado - Used by opencos.eda commands with --tool=vivado.

Contains classes for ToolVivado, and command handlers for sim, elab, synth, build,
upload, flist, open, proj.
'''

# pylint: disable=R0801 # (setting similar, but not identical, self.defines key/value pairs)
# pylint: disable=too-many-lines

import os
import re
import shlex

from pathlib import Path

from opencos import util, eda_base
from opencos.commands import sim, CommandSim, CommandSynth, CommandProj, CommandBuild, \
    CommandFList, CommandUpload, CommandOpen, CommandWaves
from opencos.eda_base import Tool
from opencos.files import safe_shutil_which, get_source_files_paths
from opencos.util import Colors
from opencos.utils import docker_checks
from opencos.utils.docker_checks import docker_ok


class ToolVivado(Tool):
    '''ToolVivado used by opencos.eda for --tool=vivado'''

    _TOOL = 'vivado'
    _EXE = 'vivado'

    vivado_year = None
    vivado_release = None
    vivado_base_path = ''
    vivado_exe = ''

    # use if config.tools.TOOL.docker-default-image is missing:
    DEFAULT_DOCKER_IMAGE = 'siliconbootcamp/xsim-synth-2024:stable'

    def __init__(self, config: dict):
        super().__init__(config=config) # calls self.get_versions()
        self.args.update({
            'part': 'xcu200-fsgd2104-2-e',
            'add-glbl-v': False,
        })
        self.args_help.update({
            'part': 'Device used for commands: synth, build.',
            'add-glbl-v': (
                '(for simulation) add glbl.v to filelist, "glbl" to sim libraries, and "glbl"'
                ' as an additional top level hier.'
            ),
        })


    def get_versions(self, **kwargs) -> str: # pylint: disable=too-many-branches
        if self._VERSION:
            return self._VERSION

        path = safe_shutil_which(self._EXE)
        if not path:
            # but can we run via docker? Note docker isn't required, this is a fall-back
            if not docker_ok():
                if self.get_version_allow_certain_command_if_tool_not_installed():
                    return self._VERSION
                self.error("Vivado not in path, need to install or add to $PATH",
                           f"(looked for '{self._EXE}')")
                return self._VERSION

            self.vivado_exe = 'vivado'
            self.vivado_base_path = '' # clear these so they work in docker image
            self.no_exe_but_has_docker = True

            self.get_set_default_docker_image()

            if not docker_checks.check_image_available(image=self.DEFAULT_DOCKER_IMAGE):
                # Be nice to eda --help, don't call docker run to get versioning if
                # the image hasn't been pulled.
                self._VERSION = 'IMAGE-NOT-PULLED'
                return self._VERSION

        else:
            self.vivado_exe = path
            self.vivado_base_path, _ = os.path.split(path)


        version_path = ''
        if path:
            version_path = path
            xilinx_vivado = os.environ.get('XILINX_VIVADO')

            # Since get_versions() doesn't have a quiet mode, and it's called as part
            # of opencos.eda.init_config(..), let's only show an information message
            # if XILINX_VIVADO env is set and it doesn't match the Path/bin/vivado[.bat|.exe]
            if xilinx_vivado and \
               os.path.abspath(os.path.join(xilinx_vivado, 'bin')) != \
                   os.path.abspath(os.path.dirname(self.vivado_exe)):
                util.info(
                    "environment for XILINX_VIVADO is set, and doesn't match the vivado path:",
                    f"XILINX_VIVADO={xilinx_vivado} EXE PATH={self.vivado_exe}"
                )

        elif docker_ok():
            # Try to get the version from the docker image
            env_dict = docker_checks.get_docker_env(cmd_des_obj=self)
            version_ret = docker_checks.run_docker_for_version_info(
                image=self.DEFAULT_DOCKER_IMAGE,
                env_dict=env_dict,
                commands=['-c', 'which vivado']
            )
            if version_ret:
                version_path = version_ret
            else:
                # try from the env path?
                env_paths = env_dict.get('PATH', '').split(':')
                for p in env_paths:
                    if 'Vivado' in p:
                        version_path = p

        # Get version based on install path name. Calling vivado -verison is too slow.
        util.debug(f"vivado path = {version_path}")
        m = re.search(r'(\d\d\d\d)\.(\d)', version_path)
        if m:
            version = m.group(1) + '.' + m.group(2)
            self._VERSION = version
        else:
            version = ''
            util.warn_once("Vivado path doesn't specify version, expecting (dddd.d)",
                           warn_name='vivado_path_no_version')
            if not path and docker_ok():
                # Note - we'll let this try again, because docker might not have the --docker-image
                # arg parsed yet.
                pass
            else:
                version = 'UNKNOWN'

        if version:
            numbers_list = version.split('.')
            self.vivado_year = int(numbers_list[0])
            self.vivado_release = int(numbers_list[1])
            self.vivado_version = float(numbers_list[0] + '.' + numbers_list[1])
        else:
            util.warn_once(f"Vivado version not found, vivado path = {version_path}",
                           warn_name='vivado_version_not_found')
        return self._VERSION


    def set_tool_defines(self) -> None:
        self.defines['OC_TOOL_VIVADO'] = None
        if not self.vivado_year or not self.vivado_release:
            return
        def_year_release = f'OC_TOOL_VIVADO_{self.vivado_year:04d}_{self.vivado_release:d}'
        self.defines[def_year_release] = None

        # Code can be conditional on Vivado versions and often keys of "X or older" ...
        versions = ['2021.1', '2021.2', '2022.1', '2022.2', '2023.1', '2023.2',
                    '2024.1', '2024.2']
        for ver in versions:
            float_ver = float(ver)
            str_ver = str(float_ver).replace('.', '_')
            if self.vivado_version <= float_ver:
                self.defines[f'OC_TOOL_VIVADO_{str_ver}_OR_OLDER'] = None
            if self.vivado_version >= float_ver:
                self.defines[f'OC_TOOL_VIVADO_{str_ver}_OR_NEWER'] = None

        # Older Vivado's don't correctly compare types in synthesis (xsim seems OK)
        if self.vivado_version < 2023.2:
            self.defines['OC_TOOL_BROKEN_TYPE_COMPARISON'] = None

        util.debug(f"Setup tool defines: {self.defines}")


    def get_vivado_tcl_verbose_arg(self) -> str:
        '''Returns a common Vivado tcl arg str (-verbose, -quiet, or both/none)'''
        v = "" # v = verbose tcl arg we'll add to many tcl commands.
        if util.args.get('verbose', ''):
            v += " -verbose"
        elif util.args.get('quiet', ''):
            v += " -quiet"
        return v


class CommandWavesVivado(CommandWaves, ToolVivado):
    '''Handler for: eda waves --tool=vivado'''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolVivado.__init__(self, config=self.config)

    def do_it(self) -> None:
        tcl_name = self.wave_file + '.eda_waves.tcl'
        with open(tcl_name, 'w', encoding='utf-8') as fo :
            print('current_fileset', file=fo)
            print(f'open_wave_database {self.wave_file}', file=fo)

        command_list = [self.vivado_exe, '-source', tcl_name]
        self.exec(os.path.dirname(self.wave_file), command_list)


class CommandSimVivado(CommandSim, ToolVivado):
    '''CommandSimVivado is a command handler for: eda sim --tool=vivado, uses xvlog, xelab, xsim

    Note that we attempt to run a generated .tcl script within vivado, that will perform the
    3-step compile/elaborate/simulate steps'''

    library_map_supported = True


    def __init__(self, config: dict):
        CommandSim.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args.update({
            'gui':        False,
            'tcl-file':   'sim.tcl',
            'fpga':       '',
            'add-glbl-v': False,
            'all-sv':     False,
            'docker-image': self.DEFAULT_DOCKER_IMAGE
        })
        self.args_help.update({
            'gui':        'Run Vivado XSim in GUI mode',
            'tcl-file':   'name of TCL file to be created for XSim',
            'fpga':       'FPGA device name, can be used for various Xilinx IP or XCIs',
            'add-glbl-v': (
                'Use the glbl.v in xvlog, xelab for this version of Vivado. Adds glbl.v to'
                ' filelist(s), and add "glbl" as an additional top level hier to xelab.'
            ),
            'library-map': (
                'list arg (can set multiple) with values LibName:Path, for example:'
                ' --library-map=unisim:path/to/custom_lib_name.lib'
                ' The Path search is relative to the tool exe, one level above the tool exe,'
                ' or can be absolute path.'
            )
        })

        self.sim_libraries = self.tool_config.get('sim-libraries', [])

        self.vivado_tcl = {
            'xvlog': [],
            'xelab': [],
            'xsim': [],
            'exe_list': [],
            'check_logs': [],
        }

        # Note this is the syntax to have your Vivado tcl print to
        # stdout, which we do for xvlog, xelab always, and even for xsim if in
        # --gui or not GUI
        self.tcl_exec_pipe = ['>@stdout', '2>@stderr']


    def set_tool_defines(self):
        ToolVivado.set_tool_defines(self)

    # We do not override CommandSim.do_it(), CommandSim.check_logs_for_errors(...)

    def prepare_compile(self) -> None: # pylint: disable=too-many-branches
        self.set_tool_defines()
        self.update_library_map()

        self.docker_arg_set = self.args['docker-run']
        if self.no_exe_but_has_docker and not self.docker_arg_set:
            self.error('"vivado" was not found in PATH, but can be run with --docker-run',
                       f'and optionally --docker-image={self.DEFAULT_DOCKER_IMAGE}')
            return

        if self.docker_arg_set:
            if docker_ok():
                # If this is docker, we cannot use the full path for any exe,
                # we have to simply use 'vivado' in the shell scripts that the docker image
                # will run.
                self.vivado_exe = 'vivado'
                util.info(
                    f'Running with --docker-run --docker-image={self.args["docker-image"]}',
                    color=Colors.byellow
                )
            else:
                self.error(docker_checks.get_error_str_cant_run_docker())

        # Don't use the return values, these will set values in self.vivado_tcl:
        self.get_compile_command_lists()
        self.get_elaborate_command_lists()
        self.get_simulate_command_lists()

        # We will always run a generated .tcl scirpt from vivado as:
        # one command line call for:
        #    vivado -mode batch -source all_vivado.tcl
        # So we have to create all_vivado.tcl based on our commands that we saved
        # in self.vivado_tcl:
        tclfname = os.path.abspath(
            os.path.join(self.args['work-dir'], 'all_vivado.tcl')
        )
        with open(tclfname, 'w', encoding='utf-8' ) as ftcl:
            for line in self.vivado_tcl['xvlog']:
                if self.args['stop-before-compile']:
                    ftcl.write('# ')
                ftcl.write(line)
                ftcl.write('\n\n')
            for line in self.vivado_tcl['xelab']:
                if any(self.args[x] for x in ('stop-before-compile', 'stop-after-compile')):
                    ftcl.write('# ')
                ftcl.write(line)
                ftcl.write('\n\n')
            for line in self.vivado_tcl['xsim']:
                if any(self.args[x] for x in ('stop-before-compile', 'stop-after-compile',
                                              'stop-after-elaborate')):
                    ftcl.write('# ')
                ftcl.write(line)
                ftcl.write('\n\n')

        # We will alwyas run this in -mode batch, the xsim call in
        # all_vivado.tcl will have -gui in it:
        self.vivado_tcl['exe_list'] = [
            self.vivado_exe, '-mode', 'batch', '-source', 'all_vivado.tcl'
        ]

        # Since we run this in a single vivado call, we don't get the automatic
        # log checking from sim.run_commands_check_logs(..), so we need to remember
        # which logs to check:
        if self.args['stop-before-compile']:
            self.vivado_tcl['check_logs'] = []
        elif self.args['stop-after-compile']:
            self.vivado_tcl['check_logs'] = ['vivado.log', 'xvlog.log']
        elif self.args['stop-after-elaborate']:
            self.vivado_tcl['check_logs'] = ['vivado.log', 'xvlog.log', 'xelab.log']
        else:
            self.vivado_tcl['check_logs'] = ['vivado.log', 'xvlog.log', 'xelab.log', 'xsim.log']

        self.write_sh_scripts_to_work_dir(
            compile_lists=[],
            elaborate_lists=[],
            simulate_lists=[self.vivado_tcl['exe_list']],
            compile_line_breaks=False
        )

        if self.docker_arg_set:
            self.prepare_compile_docker()


    def prepare_compile_docker(self) -> None:
        '''Takes existing generated shell scripts, and saves out different ones.

        will be used in compile() and simulate() steps.

        1. Assume the existing self.vivado_tcl['exe_list'] exists (all.sh)
           a. Create docker_wrap_simulate.sh that has:
               docker run --rm --user $(id -u):$(id -g) [-v ...]  --entrypoint /bin/bash
                   [-e ...] IMAGE -c 'cd WORK-DIR && ./simulate.sh'

           b. where simulate.sh
              i. The original contents from self.vivado_tcl['exe_list'], but
                 'vivado' instead of '/path/to/vivado', which was handled
                 elsewhere looking at self.args['docker']

        3. Overwrite all.sh to call:
           - docker_wrap_simulate.sh

        4. Destroy the self.vivado_tcl['exe_list'] to the bash scripts.
        '''
        if not self.docker_arg_set:
            return

        # TODO(drew): Windows friendly? Probably not. a lot of ./ going on here.

        work_dir_abs = Path(self.args['work-dir']).resolve()
        docker_command = 'run'
        docker_env_dict = docker_checks.get_docker_env(cmd_des_obj=self)
        docker_run_base_cmd = docker_checks.get_docker_run_command(
            image=self.args['docker-image'],
            docker_command=docker_command,
            env_dict=docker_env_dict,
            ro_volumes=get_source_files_paths(cmd_des_obj=self, minimal_root_paths=True),
            rw_volumes=[work_dir_abs],
            threads=self.args['threads']
        )

        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='docker_wrap_simulate.sh',
            command_lists=[
                util.ShellCommandList(docker_run_base_cmd + [
                    '-c', f'\'cd {work_dir_abs} && ./simulate.sh\'',
                ], tee_fpath='sim.log')
            ],
            line_breaks=False
        )


        all_lists = [ ] # list - of - (command-list)
        if self.has_pre_compile_dep_shell_commands:
            all_lists.append(['./pre_compile_dep_shell_commands.sh'])

        all_lists.append(['./docker_wrap_simulate.sh'])

        if self.has_post_tool_dep_shell_commands:
            all_lists.append(['./post_tool_dep_shell_commands.sh'])

        util.write_shell_command_file(
            dirpath=self.args['work-dir'], filename='all.sh',
            command_lists=all_lists, line_breaks=True
        )

        # Overwrite our command lists so simulate() work correctly.
        if self.vivado_tcl['exe_list']:
            self.vivado_tcl['exe_list'] = util.ShellCommandList(
                    ['./docker_wrap_simulate.sh'], tee_fpath='docker_wrap_simulate.log'
            )

        return


    def compile(self):
        # handled in simulate()
        return

    def elaborate(self):
        # handled in simulate()
        return

    def simulate(self):
        # even though we run one command line call (vivado -mode batch -sourc TCLFILE)
        # it still saves xvlog.log, xelab.log, xsim.log, and a (useless) vivado.log.
        # we'll have to manually check the logs for errors if this was run in --gui:
        self.run_commands_check_logs(
            [self.vivado_tcl['exe_list']], check_logs=False, log_filename='vivado.log',
            sim_step='simulate'
        )

        for log_fname in self.vivado_tcl['check_logs']:
            filename = os.path.join(self.args['work-dir'], log_fname)
            self.check_logs_for_errors(filename=filename, sim_step='simulate')


    def get_compile_command_lists(self, **kwargs) -> list:
        '''Override from sim.CommandSim - which expects a return type list.

        Since we run all in Vivado tcl, we always return an empty list,
        and instead save in self.vivado_tcl['xvlog'].
        '''
        self.set_tool_defines()

        if self.args['add-glbl-v']:
            self._add_glbl_v()

        # compile verilog
        if self.files_v:
            self.add_xvlog_commands(files_list=self.files_v, typ='v')

        # compile systemverilog
        if self.files_sv:
            self.add_xvlog_commands(files_list=self.files_sv, typ='sv')

        return []

    def process_parameters_get_list(
            self, arg_prefix: str = '-G', **kwargs
    ) -> list:
        '''Override from sim.CommandSim

        custom handler for parameters, instead of the one in sim.py
        all will be passed as -generic_top "k=v", avoids all unknown shell or bash
        behavior in windows.
        '''

        # ignore arg_prefix (use -generic_top):
        ret = []
        for k,v in self.parameters.items():
            if not isinstance(v, (int, str)):
                util.warning(
                    f'parameter {k} has value: {v}, parameters must be int/string types'
                )
            if isinstance(v, int):
                ret.extend(['-generic_top', f'"{k}={v}"'])
            else:
                v = f'{v}'.strip('\n') # stringify
                # Because we're writing to a .tcl file, \" will become ", and \\\" will become \"
                # we want \" in the final file.
                v = v.replace('"', '\\\"')
                ret.extend(['-generic_top', f'"{k}={v}"'])
        return ret


    def get_elaborate_command_lists(self, **kwargs) -> list:
        '''Override from sim.CommandSim - which expects a return type list.

        Since we run all in Vivado tcl, we always return an empty list,
        and instead save in self.vivado_tcl['xelab'].
        '''

        command_list = ['exec', 'xelab', self.args['top']]

        command_list += self.tool_config.get(
            'elab-args', '-s snapshot -timescale 1ns/1ps --stats').split()

        command_list.extend(
            self.process_parameters_get_list(arg_prefix='-generic_top ')
        )

        if self.tool_config.get('elab-waves-args', ''):
            command_list += self.tool_config.get('elab-waves-args', '').split()
        if self.args['uvm']:
            command_list.extend(['-L', 'uvm'])
        elif self.args['gui'] and self.args['waves']:
            command_list += ['-debug', 'all']
        elif self.args['gui']:
            command_list += ['-debug', 'typical']
        elif self.args['waves']:
            command_list += ['-debug', 'wave']
        if util.args['verbose']:
            command_list += ['-v', '2']
        if self.args['sim-library']:
            self.sim_libraries += self.args['sim-library'] # Add any command line libraries
        for x in self.sim_libraries:
            command_list += ['-L', x]
        if self.args['add-top-library']:
            command_list += self.args['add-top-library'] # list

        command_list += self.args['elab-args']

        # For Windows compatibility, we have some issues with command/Powershell passing args
        # so as a workaround we'll create a .tcl script xelab_vivado.tcl:
        #   exec {command_list} >@stdout 2>@stderr
        # with the caveat that it needs POSIX style paths in the vivado tclsh,
        # and we'll return the list to run it:
        # and we'll save this to self.vivado_tcl['xelab']
        command_list += self.tcl_exec_pipe
        tclfname = os.path.abspath(
            os.path.join(self.args['work-dir'], 'xelab_vivado.tcl')
        )
        line = ' '.join(command_list).replace('\n', ' ')
        self.vivado_tcl['xelab'].append(line)
        with open(tclfname, 'w', encoding='utf-8' ) as ftcl:
            ftcl.write(line)
        # return list that will go unused
        return [ ]


    def get_simulate_command_lists(self, **kwargs) -> list:
        '''Override from sim.CommandSim - which expects a return type list.

        Since we run all in Vivado tcl, we always return an empty list,
        and instead save in self.vivado_tcl['xsim'].
        '''
        # create TCL for in-simulation
        sim_tcl_name = os.path.abspath(os.path.join(self.args['work-dir'], self.args['tcl-file']))

        if self.args['waves']:
            util.artifacts.add_extension(
                search_paths=self.args['work-dir'], file_extension='wdb',
                typ='waveform', description='Vivado XSim Waveform WDB (Wave DataBase) file'
            )

        with open( sim_tcl_name, 'w', encoding='utf-8' ) as fo:
            if self.args['waves']:
                if self.args['waves-start']:
                    print(f"run {self.args['waves-start']} ns", file=fo)
                print("log_wave -recursive *", file=fo)
            print("run -all", file=fo)
            if not self.args['gui'] or self.args['test-mode']:
                print("exit", file=fo)

        sv_seed = str(self.args['seed'])

        assert isinstance(self.args["sim-plusargs"], list), \
            f'{self.target=} {type(self.args["sim-plusargs"])=} but must be list'

        # xsim uses: --testplusarg name=value
        # Note - this was problematic in Windows Powershell for --testplusarg "name=value"
        # --testplusarg "name" would work, but the only way for xsim to parse it was to
        # run it in vivado's tclsh.
        xsim_plusargs_list = []
        for x in self.args['sim-plusargs']:
            xsim_plusargs_list.append('--testplusarg')
            if x[0] == '+':
                x = x[1:]
            x = x.replace('"', '\\\"') # we have to preserve " in the value.
            xsim_plusargs_list.append(f'\"{x}\"')

        # execute snapshot
        command_list = ['exec', 'xsim']
        command_list += self.tool_config.get('simulate-args', 'snapshot --stats').split()
        if self.args['gui'] and not self.args['test-mode']:
            command_list += ['-gui']
        command_list += [
            '--tclbatch', Path(sim_tcl_name).as_posix(), # running in tclsh needs POSIX
            "--sv_seed", sv_seed
        ]
        command_list += xsim_plusargs_list
        command_list += self.args['sim-args']

        # For Windows compatibility we have some issues with command/Powershell passing args
        # so as a workaround we'll create a .tcl script xsim_vivado.tcl:
        #   exec {command_list} >@stdout 2>@stderr
        # with the caveat that it needs POSIX style paths in the vivado tclsh,
        # and we'll save this to self.vivado_tcl['xsim']
        command_list += self.tcl_exec_pipe

        tclfname = os.path.abspath(
            os.path.join(self.args['work-dir'], 'xsim_vivado.tcl')
        )
        line = ' '.join(command_list).replace('\n', ' ')
        self.vivado_tcl['xsim'].append(line)
        with open(tclfname, 'w', encoding='utf-8' ) as ftcl:
            ftcl.write(line)

        # Need to return list, will go unused.
        return []

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []


    def add_xvlog_commands( # pylint: disable=too-many-branches
            self, files_list: list, typ: str = 'sv'
    ) -> None:
        '''Returns None, because we'll save the results in self.vivado_tcl['xvlog'].

        Vivado still treats .v files like Verilog-2001, so we split xvlog into .v and .sv sections,
        as two entries in self.vivado_tcl['xvlog'] if self.args['all-sv'] is False.
        '''


        if not files_list:
            return

        # For Windows compatibility we have some issues with command/Powershell passing args
        # so as a workaround we'll create a .tcl script xvlog_sv_vivado.tcl:
        #   exec {command_list} >@stdout 2>@stderr
        # with the caveat that it needs POSIX style paths in the vivado tclsh,
        # and we'll save this to self.vivado_tcl['xvlog']
        command_list = ['exec', 'xvlog']
        if typ == 'sv':
            command_list.append('-sv')
            if self.args['uvm']:
                command_list.extend(['-L', 'uvm'])
        command_list += self.tool_config.get('compile-args', '').split()
        if util.args['verbose']:
            command_list += ['-v', '2']

        # If there were any external .lib files from --library-map args, add those now:
        command_list.extend(self._get_xvlog_library_map_list())

        for value in self.incdirs:
            command_list.append('-i')
            command_list.append(Path(value).as_posix())

        if self.args['ext-defines-sv-fname']:
            self.create_ext_defines_sv()
        else:
            for key, value in self.defines.items():
                command_list.append('-d')
                if value is not None:
                    # Because we're writing to a .tcl file, \" will become ", and \\\" will become
                    # \". We want \" in the final file. Parameters need to act the same way as
                    # defines:
                    value = f'{value}'.replace('"', '\\\"')

                if value is None:
                    command_list.append(key)
                else:
                    command_list.append(f"\"{key}={value}\"")

        command_list += self.args['compile-args']

        # Convert these to POSIX (even though we might be in Windows) b/c that's what Vivado
        # tclsh needs
        command_list += [Path(fpath).as_posix() for fpath in files_list]
        command_list += self.tcl_exec_pipe

        tclfname = os.path.abspath(
            os.path.join(self.args['work-dir'], f'xvlog_{typ}_vivado.tcl')
        )

        # Make a one-liner from command line, save in self.vivado_tcl['xvlog']:
        line = ' '.join(command_list).replace('\n', ' ')
        self.vivado_tcl['xvlog'].append(line)
        with open(tclfname, 'w', encoding='utf-8' ) as ftcl:
            ftcl.write(line)
            ftcl.write('\n')


    def _add_glbl_v(self):
        '''Adds glbl.v from Vivado's install path to self.files_v'''
        glbl_v = self.vivado_base_path.replace(
            'bin', os.path.join('data', 'verilog', 'src', 'glbl.v')
        )
        if any(x and os.path.split(x)[1] == 'glbl.v' for x in self.files_v):
            util.warning(f'--add-glbl-v: Not adding {glbl_v=} b/c glbl.v already in',
                         f'{self.files_v=}')
        elif not os.path.exists(glbl_v):
            self.error(f"Could not find file {glbl_v=}")
        else:
            self.files_v.insert(0, glbl_v)
            # add to self.args['add-top-library']
            if 'glbl' not in self.args['add-top-library']:
                self.args['add-top-library'].append('glbl')

    def _get_xvlog_library_map_list(self) -> list:
        '''Returns list of lines used in vlog step for --library-map args

        Note that Vivado does not do library name re-mapping, the name comes
        from the .lib file internals.
        '''

        lines = []
        for _, lib_file_path in self.library_map.items():
            lines.append(f'-libmap {lib_file_path}')
        return lines


    def artifacts_add(self, name: str, typ: str, description: str) -> None:
        '''Override from Command.artifacts_add, so we can catch known file

        names to make their typ/description better, such as CommandSim using
        sim.log or compile.log
        '''
        _, leafname = os.path.split(name)
        if leafname == 'vivado.log':
            description = 'Vivado XSim log from stdout/stderr'
        elif leafname == 'xsim.log':
            description = 'Vivado XSim simulation step (3/3) log from stdout/stderr'
        elif leafname == 'xelab.log':
            description = 'Vivado XSim elaboration step (2/3) log from stdout/stderr'
        elif leafname == 'xvlog.log':
            description = 'Vivado XSim compile step (1/3) log from stdout/stderr'

        super().artifacts_add(name=name, typ=typ, description=description)


class CommandElabVivado(CommandSimVivado):
    '''CommandElabVivado is a command handler for: eda elab --tool=vivado, uses xvlog, xelab'''
    command_name = 'elab'
    def __init__(self, config: dict):
        CommandSimVivado.__init__(self, config)
        # add args specific to this tool
        self.args['stop-after-elaborate'] = True


class CommandLintVivado(CommandSimVivado):
    '''CommandLintVivado is a command handler for: eda lint --tool=vivado, uses xvlog'''
    command_name = 'lint'
    def __init__(self, config: dict):
        CommandSimVivado.__init__(self, config)
        # add args specific to this tool
        self.args['stop-after-compile'] = True
        self.args['stop-after-elaborate'] = True


class CommandSynthVivado(CommandSynth, ToolVivado):
    '''CommandSynthVivado is a command handler for: eda synth --tool=vivado'''
    def __init__(self, config: dict):
        CommandSynth.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args['gui'] = False
        self.args['tcl-file'] = "synth.tcl"
        self.args['xdc'] = ""
        self.args['fpga'] = ""
        self.args['all-sv'] = False

    def do_it(self) -> None:
        CommandSynth.do_it(self)

        if self.is_export_enabled():
            return

        # create TCL
        tcl_file = os.path.abspath(
            os.path.join(self.args['work-dir'], self.args['tcl-file'])
        )

        self.write_tcl_file(tcl_file=tcl_file)

        # execute Vivado
        command_list = [
            self.vivado_exe, '-mode', 'batch', '-source', tcl_file,
            '-log', f"{self.args['top']}.synth.log"
        ]
        if not util.args['verbose']:
            command_list.append('-notrace')
        self.exec(self.args['work-dir'], command_list)


    def write_tcl_file( # pylint: disable=too-many-locals,too-many-branches
            self, tcl_file: str
    ) -> None:
        '''Writes synthesis capable Vivado tcl file to filepath 'tcl_file'.'''

        # TODO(drew): This method needs to be broken up to avoid the pylint
        # waivers.

        v = self.get_vivado_tcl_verbose_arg()

        defines = ""
        for key, value in self.defines.items():
            defines += (f"-verilog_define {key}" + (" " if value is None else f"={value} "))
        parameters = ' '.join(
            sim.parameters_dict_get_command_list(params=self.parameters, arg_prefix='-generic ')
        )
        incdirs = ' '.join([f'-include_dirs {Path(x).as_posix()}' for x in self.incdirs])
        flatten = ""
        if self.args['flatten-all']:
            flatten = "-flatten_hierarchy full"
        elif self.args['flatten-none']:
            flatten = "-flatten_hierarchy none"

        tcl_lines = []
        # Note - if we're in Windows, the vivado tcl that processes these lines expects POSIX
        # paths, not command/Powershell paths. so we will use Path(str).as_posix()
        for f in self.files_v:
            tcl_lines.append(f"read_verilog {Path(f).as_posix()}")
        for f in self.files_sv:
            tcl_lines.append(f"read_verilog -sv {Path(f).as_posix()}")
        for f in self.files_vhd:
            tcl_lines.append(f"add_file {Path(f).as_posix()}")

        part = self.args['part']
        top = self.args['top']

        default_xdc = False
        if self.args['xdc'] != "":
            xdc_file = os.path.abspath(self.args['xdc'])
        elif self.files_sdc:
            # Use files from DEPS target or command line.
            xdc_file = ''
        else:
            default_xdc = True
            xdc_file = os.path.abspath(os.path.join(self.args['work-dir'],
                                                    "default_constraints.xdc"))


        tcl_lines += [
            f"create_fileset -constrset constraints_1 {v}",
        ]
        for _file in self.files_sdc:
            # NOTE - sdc files cannot (yet) be attached to other modules.
            tcl_lines += [
                f"add_files -fileset constraints_1 {Path(_file).as_posix()} {v}",
            ]
        if xdc_file:
            tcl_lines += [
                f"add_files -fileset constraints_1 {Path(xdc_file).as_posix()} {v}",
            ]
        tcl_lines += [
            "# FIRST PASS -- auto_detect_xpm",
            (
                f"synth_design -part {part} -rtl -rtl_skip_ip -rtl_skip_constraints"
                f" -no_timing_driven -no_iobuf -top {top} {incdirs} {defines} {parameters} {v}"
            ),
            f"auto_detect_xpm {v} ",
            f"synth_design -no_iobuf -part {part} {flatten} -constrset constraints_1 " \
            + f"-top {top} {incdirs} {defines} {parameters} {v}",
            f"write_verilog -force {top}.vg {v}",
            f"report_utilization -file {top}.flat.util.rpt {v}",
            f"report_utilization -file {top}.hier.util.rpt {v} -hierarchical " \
            + "-hierarchical_depth 20",
            f"report_timing -file {top}.timing.rpt {v}",
            f"report_timing_summary -file {top}.summary.timing.rpt {v}",
            f"report_timing -from [all_inputs] -file {top}.input.timing.rpt {v}",
            f"report_timing -to [all_outputs] -file {top}.output.timing.rpt {v}",
            "report_timing -from [all_inputs] -to [all_outputs] " \
            + f"-file {top}.through.timing.rpt {v}",
            "set si [get_property -quiet SLACK [get_timing_paths -max_paths 1 -nworst 1 -setup " \
            + "-from [all_inputs]]]",
            "set so [get_property -quiet SLACK [get_timing_paths -max_paths 1 -nworst 1 -setup " \
            + "-to [all_outputs]]]",
            f"set_false_path -from [all_inputs] {v}",
            f"set_false_path -to [all_outputs] {v}",
            "set sf [get_property -quiet SLACK [get_timing_paths -max_paths 1 -nworst 1 -setup]]",
            "if { ! [string is double -strict $sf] } { set sf 9999 }",
            "if { ! [string is double -strict $si] } { set si 9999 }",
            "if { ! [string is double -strict $so] } { set so 9999 }",
            "puts \"\"",
            "puts \"*** ****************** ***\"",
            "puts \"***                    ***\"",
            "puts \"*** SYNTHESIS COMPLETE ***\"",
            "puts \"***                    ***\"",
            "puts \"*** ****************** ***\"",
            "puts \"\"",
            "puts \"** AREA **\"",
            "report_utilization -hierarchical",
            "puts \"** TIMING **\"",
            "puts \"\"",
        ]

        if default_xdc:
            tcl_lines += [
                f"puts \"(Used default XDC: {Path(xdc_file).as_posix()})\"",
                f"puts \"DEF CLOCK NS  : [format %.3f {self.args['clock-ns']}]\"",
                f"puts \"DEF IDELAY NS : [format %.3f {self.args['idelay-ns']}]\"",
                f"puts \"DEF ODELAY NS : [format %.3f {self.args['odelay-ns']}]\"",
            ]
        else:
            tcl_lines += [
                f"puts \"(Used provided XDC: {Path(xdc_file).as_posix()})\"",
            ]
        tcl_lines += [
            "puts \"\"",
            "puts \"F2F SLACK     : [format %.3f $sf]\"",
            "puts \"INPUT SLACK   : [format %.3f $si]\"",
            "puts \"OUTPUT SLACK  : [format %.3f $so]\"",
            "puts \"\"",
        ]

        if default_xdc:
            self.write_default_xdc(xdc_file=xdc_file)

        with open( tcl_file, 'w', encoding='utf-8' ) as ftcl:
            ftcl.write('\n'.join(tcl_lines))


    def write_default_xdc(self, xdc_file: str) -> None:
        '''Writes a default XDC file to filepath 'xdc_file'.'''

        xdc_lines = []
        util.info("Creating default constraints: clock:",
                  f"{self.args['clock-name']}, {self.args['clock-ns']} (ns),",
                  f"idelay:{self.args['idelay-ns']}, odelay:{self.args['odelay-ns']}")

        clock_name = self.args['clock-name']
        period = self.args['clock-ns']
        name_not_equal_clocks_str = f'NAME !~ "{clock_name}"'

        xdc_lines += [
            f"create_clock -add -name {clock_name} -period {period} [get_ports " \
            + "{" + clock_name + "}]",
        ]
        xdc_lines += [
            f"set_input_delay -max {self.args['idelay-ns']} -clock {clock_name} " +
            "[get_ports * -filter {DIRECTION == IN && " \
            + name_not_equal_clocks_str + "}]",
        ]
        xdc_lines += [
            f"set_output_delay -max {self.args['odelay-ns']} -clock {clock_name} " +
            "[get_ports * -filter {DIRECTION == OUT}]"
        ]
        with open( xdc_file, 'w', encoding='utf-8' ) as fxdc:
            fxdc.write('\n'.join(xdc_lines))



class CommandProjVivado(CommandProj, ToolVivado):
    '''CommandProjVivado is a command handler for: eda proj --tool=vivado'''

    def __init__(self, config: dict):
        CommandProj.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args['gui'] = True
        self.args['oc-vivado-tcl'] = True
        self.args['tcl-file'] = "proj.tcl"
        self.args['xdc'] = ""
        self.args['board'] = ""
        self.args['all-sv'] = False

    def do_it(self):
        # add defines for this job
        self.set_tool_defines()
        self.write_eda_config_and_args()

        oc_root = util.get_oc_root()

        # create TCL
        tcl_file = os.path.abspath(os.path.join(self.args['work-dir'], self.args['tcl-file']))
        v = self.get_vivado_tcl_verbose_arg()

        incdirs = " ".join([Path(x).as_posix() for x in self.incdirs])
        defines = ""
        for key, value in self.defines.items():
            defines += (f"{key} " if value is None else f"{key}={value} ")

        tcl_lines = [
            f"create_project {self.args['top']}_proj {Path(self.args['work-dir']).as_posix()} {v}"
        ]

        if self.args['oc-vivado-tcl'] and oc_root:
            tcl_lines += [
                f"source \"{oc_root}/boards/vendors/xilinx/oc_vivado.tcl\" -notrace"
            ]

        if self.args['board']:
            tcl_lines += [
                f"set_property board_part {self.args['board']} [current_project]"
            ]

        tcl_lines += [
            f"set_property include_dirs {{{incdirs}}} [get_filesets sources_1]",
            f"set_property include_dirs {{{incdirs}}} [get_filesets sim_1]",
            f"set_property verilog_define {{{defines}}} [get_filesets sources_1]",
            f"set_property verilog_define {{SIMULATION {defines}}} [get_filesets sim_1]",
            "set_property -name {STEPS.SYNTH_DESIGN.ARGS.MORE OPTIONS} -value " \
            + "{{-verilog_define SYNTHESIS}} -objects [get_runs synth_1]",
            "set_property {xsim.simulate.runtime} {10ms} [get_filesets sim_1]",
            "set_property {xsim.simulate.log_all_signals} {true} [get_filesets sim_1]",
        ]

        for f in self.files_v + self.files_sv + self.files_vhd:
            # TODO(drew): automatically adding some files to sim_1 vs sources_1 should be
            # configurable in eda_config_defaults.yml or via some custom eda arg.
            if any(x in f for x in ['/sim/', '/tests/']):
                fileset = "sim_1"
            else:
                fileset = "sources_1"
            tcl_lines += [
                f"add_files -norecurse {Path(f).as_posix()} -fileset [get_filesets {fileset}]"
            ]

        tcl_lines += [
            f"set_property top {self.args['top']} [get_filesets sim_1]"
        ]
        with open( tcl_file, 'w', encoding='utf-8' ) as fo:
            fo.write('\n'.join(tcl_lines))

        # execute Vivado
        command_list = [
            self.vivado_exe, '-mode', 'gui', '-source', tcl_file,
            '-log', f"{self.args['top']}.proj.log"
        ]
        if not util.args['verbose']:
            command_list.append('-notrace')
        self.exec(self.args['work-dir'], command_list)
        util.info(f"Project run done, results are in: {self.args['work-dir']}")


class CommandBuildVivado(CommandBuild, ToolVivado):
    '''CommandBuildVivado is a command handler for: eda build --tool=vivado'''

    def __init__(self, config: dict):
        CommandBuild.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args['gui'] = False
        self.args['fpga'] = ""
        self.args['proj'] = False
        self.args['resynth'] = False
        self.args['reset'] = False
        self.args['all-sv'] = False

    def do_it(self):
        # add defines for this job
        self.set_tool_defines()
        self.write_eda_config_and_args()

        # create FLIST
        flist_file = os.path.abspath(os.path.join(self.args['work-dir'],'build.flist'))
        util.debug(f"CommandBuildVivado: top={self.args['top']} target={self.target}",
                   f"design={self.args['design']}")

        eda_path = eda_base.get_eda_exec('flist')
        command_list = [
            eda_path, 'flist',
            '--no-default-log',
            '--tool=' + self.args['tool'],
            '--force',
            '--out=' + flist_file,
            '--no-quote-define',
            '--quote-define-value',
            '--escape-define-value',
            '--no-equal-define',
            '--bracket-quote-path',
            # on --prefix- items, use shlex.quote(str) so spaces work with subprocess shell=False:
            '--prefix-incdir=' + shlex.quote("oc_set_project_incdir "),
            '--prefix-define=' + shlex.quote("oc_set_project_define "),
            '--prefix-sv=' + shlex.quote("add_files -norecurse "),
            '--prefix-v=' + shlex.quote("add_files -norecurse "),
            '--prefix-vhd=' + shlex.quote("add_files -norecurse "),
        ]

        # create an eda.flist_input.f that we'll pass to flist:
        with open(os.path.join(self.args['work-dir'], 'eda.flist_input.f'),
                  'w', encoding='utf-8') as f:
            f.write('\n'.join(self.files_v + self.files_sv + self.files_vhd + ['']))

        command_list.append('--input-file=eda.flist_input.f')

        for key,value in self.defines.items():
            if value is None:
                command_list += [ f"+define+{key}" ]
            else:
                command_list += [ shlex.quote(f"+define+{key}={value}") ]
        cwd = util.getcwd()
        util.debug(f"CommandBuildVivado: {cwd=}")


        # Write out a .sh command, but only for debug, it is not run.
        command_list = util.ShellCommandList(command_list, tee_fpath='run_eda_flist.log')
        util.write_shell_command_file(dirpath=self.args['work-dir'], filename='run_eda_flist.sh',
                                      command_lists=[command_list], line_breaks=True)

        ###self.exec(cwd, command_list, tee_fpath=command_list.tee_fpath)
        # Run this from work-dir
        self.exec(work_dir=self.args['work-dir'], command_list=command_list,
                  tee_fpath=command_list.tee_fpath)

        if self.args['job-name'] == "":
            self.args['job-name'] = self.args['design']
        project_dir = 'project.'+self.args['job-name']

        # launch Vivado
        command_list = [self.vivado_exe]
        command_list += [
            '-mode',
            'gui' if self.args['gui'] and not self.args['test-mode'] else 'batch',
            '-log', os.path.join(self.args['work-dir'], self.args['top'] + '.build.log')
        ]
        if not util.args['verbose']:
            command_list.append('-notrace')
        command_list += [
            '-source', self.args['build-script'],
            '-tclargs', project_dir,
            # these must come last, all after -tclargs get passed to build-script
            flist_file,
        ]
        if self.args['proj']:
            command_list += ['--proj']
        if self.args['resynth']:
            command_list += ['--resynth']
        if self.args['reset']:
            command_list += ['--reset']

        # Write out a .sh command, but only for debug, it is not run.
        command_list = util.ShellCommandList(command_list, tee_fpath=None)
        util.write_shell_command_file(dirpath=self.args['work-dir'], filename='run_vivado.sh',
                                      command_lists=[command_list], line_breaks=True)

        if self.args['stop-before-compile']:
            util.info(f"--stop-before-compile set: scripts in : {self.args['work-dir']}")
            return

        # Run this from current working dir (not work-dir)
        self.exec(cwd, command_list, tee_fpath=command_list.tee_fpath)
        util.info(f"Build done, results are in: {self.args['work-dir']}")


class CommandFListVivado(CommandFList, ToolVivado):
    '''CommandFlistVivado is a command handler for: eda flist --tool=vivado

    This is more synthesis and projects, not for Vivado XSim. We'll log
    warnings if any plusargs or parameters were set, because they will
    not be emitted.
    '''

    def __init__(self, config: dict):
        CommandFList.__init__(self, config=config)
        ToolVivado.__init__(self, config=self.config)
        self.args.update({
            'all-sv': False,
            # synth/project style Flist, can't emit these:
            'emit-parameter': False,
            'emit-plusargs': False,
        })

    def get_flist_plusargs_list(self) -> list:
        '''Overriden from CommandFList.'''
        if self.args['unprocessed-plusargs']:
            util.warning('Command "flist" for --tool=vivado is not intended for simulation',
                         'and plusargs were present. They will NOT be included in the flist:',
                         f'{self.args["unprocessed-plusargs"]}')

        return []

    def get_flist_parameter_list(self) -> list:
        '''Overriden from CommandFList.'''
        if self.parameters:
            util.warning('Command "flist" for --tool=vivado is not intended for simulation',
                         'and parameters were present. They will NOT be included in the flist:',
                         f'{self.parameters}')

        return []


class CommandUploadVivado(CommandUpload, ToolVivado):
    '''CommandUploadVivado is a command handler for: eda upload --tool=vivado'''

    SUPPORTED_BIT_EXT = ['.bit']

    def __init__(self, config: dict):
        CommandUpload.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args.update({
            'gui': False,
            'list-usbs': False,
            'list-devices': False,
            'usb': -1,
            'device': -1,
            'host': "localhost",
            'port': 3121,
            'all-sv': False,
            'tcl-file': "eda_upload.tcl",
            'log-file': "eda_upload.log",
            'test-mode': False,
        })

        # TODO(drew): Complete self.args_help.update({...})
        self.args_help.update({
            'bitfile': 'BIT file to upload (auto-detected if not specified)',
            'list-bitfiles': 'List available BIT files',
        })


    def do_it(self): # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        '''
        Note this is called directly by opencos.commands.CommandUpload, based
         on which bitfile(s) were found, or if --tool=vivado was set

        We do not need to handle --list-bitfiles, was handled by CommandUpload.
        '''

        # add defines for this job
        self.set_tool_defines()
        self.write_eda_config_and_args()

        bitfile = None
        if self.args['bitfile']:
            if os.path.isfile(self.args['bitfile']):
                bitfile = self.args['bitfile']

        # self.bitfiles was already set by CommandUpload.process_tokens()
        if len(self.bitfiles) == 1:
            bitfile = self.bitfiles[0]


        # Auto-discover bitfile logic (for when we have no bitfile, and we
        # weren't called just to listdevice/listusb)
        if not bitfile and not self.args['list-devices'] and not self.args['list-usbs']:

            # CommandUpload already displayed them, and exited on --list-bitfiles.
            if len(self.bitfiles) > 1:
                util.warning("Too many matches to continue without adding search terms,",
                             "upload not performed.")
                if not self.args['test-mode']:
                    self.error('Upload not performed, multiple bit files found, please add',
                               'search terms...')
                return

            self.error("Failed to find a matching bitfile")
            return

        # ── Generate TCL script ───────────────────────────────────────────────────
        script_file = Path(self.args['tcl-file'])
        log_file    = Path(self.args['log-file'])

        try:
            with script_file.open("w", encoding="utf-8") as fout:
                w = fout.write   # local alias (brevity)

                w('open_hw_manager\n')
                w(f'connect_hw_server -url {self.args["host"]}:{self.args["port"]}\n')
                w('refresh_hw_server -force_poll\n')

                w('set hw_targets [get_hw_targets -quiet */xilinx_tcf/Xilinx/*]\n')
                w('set num_targets [llength $hw_targets]\n')
                if self.args['list-usbs']:
                    w('puts "\\[INFO\\] OC_LOAD_BITFILE TCL $num_targets USB targets found"\n')
                    w('for {set u 0} {$u < $num_targets} {incr u} {\n')
                    w('  puts "\\[INFO\\] OC_LOAD_BITFILE TCL: USB $u : [lindex $hw_targets $u]"\n')
                    w('}\n')
                    w('if { $num_targets > 1} {\n')
                    w('  set maxusb [expr $num_targets - 1]\n')
                    w('  puts "\\[INFO\\] OC_LOAD_BITFILE TCL: With >1 target you need to specify'
                      '--usb <n> where <n> is 0-$maxusb"\n')
                    w('}\n')
                w('if { $num_targets == 0 } {\n')
                w('  puts "\\[ERROR\\] OC_LOAD_BITFILE TCL: No HW_targets found!"\n')
                w('  exit\n}\n')

                if self.args['usb'] == -1:
                    w('if { $num_targets == 1 } {\n')
                    w('  puts "\\[INFO\\] OC_LOAD_BITFILE TCL: Defaulting to USB #0, since there is'
                      'only one device"\n')
                    w('  set usb 0\n} else {\n')
                    if not self.args['list-usbs']:
                        w('  set maxusb [expr $num_targets - 1]\n')
                        w('  puts "\\[ERROR\\] OC_LOAD_BITFILE TCL: Need --usb <n> argument, <n>'
                          'being 0-$maxusb, use --list-usbs if needed"\n')
                    w('  exit\n}\n')
                else:
                    w(f'set usb {self.args["usb"]}\n')

                w('if { $num_targets <= $usb } {\n')
                w('  puts "\\[ERROR\\] OC_LOAD_BITFILE TCL: hw_target #$usb doesn\'t exist!!"\n')
                w('  exit\n}\n')
                w('set hw_target [lindex $hw_targets $usb]\n')
                w('current_hw_target $hw_target\n')
                w('open_hw_target\n')
                w('refresh_hw_target\n')

                w('set hw_devices [get_hw_devices -quiet]\n')
                w('set num_devices [llength $hw_devices]\n')
                if self.args['list-devices']:
                    w('puts "\\[INFO\\] OC_LOAD_BITFILE TCL $num_devices devices found"\n')
                    w('for {set d 0} {$d < $num_devices} {incr d} {\n')
                    w('  puts "\\[INFO\\] OC_LOAD_BITFILE TCL: Device $d :'
                      '[lindex $hw_devices $d]"\n')
                    w('}\n')

                if self.args['device'] == -1:
                    w('if { $num_devices > 1 } {\n')
                    w('  if { [lindex $hw_devices 0] eq "arm_dap_0" } {\n')
                    w('    set hw_device [lindex $hw_devices 1]\n  } else {\n')
                    w('    set hw_device [lindex $hw_devices 0]\n  }\n')
                    w('}\n')
                else:
                    w(f'set hw_device [lindex $hw_devices {self.args["device"]  }]\n')

                w('puts "HW_DEVICE DID  : [get_property DID $hw_device]"\n')
                w('puts "HW_DEVICE PART : [get_property PART $hw_device]"\n')
                w('current_hw_device $hw_device\n')
                w('refresh_hw_device -update_hw_probes false -quiet $hw_device\n')

                if bitfile is not None:
                    w('set_property PROGRAM.FILE {' + Path(bitfile).as_posix() + '} $hw_device\n')
                    w('program_hw_devices [current_hw_device]\n')

                w('close_hw_target\n')
                w('quit\n')

        except OSError as exc:
            util.error(f"Cannot create {script_file}: {exc}")

        if bitfile is None:
            util.info("No bitfile provided or found")
        else:
            if os.path.isfile(bitfile):
                util.info(f"Using bitfile {bitfile}")
            else:
                util.warning(f"Using bitfile {bitfile}, which doesn't exist (or is not a file)")

        if self.args['test-mode']:
            util.info(f"test-mode set, upload skipped, {script_file=}")
            return

        # ── Execute Vivado ───────────────────────────────────────────────────────
        command_list = [
            self.vivado_exe, '-mode', 'batch', '-source', str(script_file), '-log', str(log_file)
        ]
        if not util.args['verbose']:
            command_list.append('-notrace')
        _, stdout, _ = self.exec(Path(util.getcwd()), command_list)

        # Do some log scraping
        for line in stdout.split('\n'):
            if line.startswith('WARNING:'):
                self.incr_tool_warning(count=1)
            elif line.startswith('ERROR:') or line.startswith('[ERROR]'):
                self.incr_tool_error(count=1)

        if not self.args['keep']:
            os.unlink(self.args['tcl-file'])

        self.report_tool_warn_error_counts()
        self.report_pass_fail()

        util.info("Upload done")


class CommandOpenVivado(CommandOpen, ToolVivado):
    '''CommandOpenVivado command handler class used by: eda open --tool vivado'''
    def __init__(self, config: dict):
        CommandOpen.__init__(self, config)
        ToolVivado.__init__(self, config=self.config)
        # add args specific to this tool
        self.args['gui'] = True
        self.args['file'] = False
        self.args['all-sv'] = False

    def do_it(self):
        if not self.args['file']:
            util.info("Searching for project...")
            found_file = False
            all_files = []
            for root, _, files in os.walk("."):
                for file in files:
                    if file.endswith(".xpr"):
                        found_file = os.path.abspath(os.path.join(root,file))
                        util.info(f"Found project: {found_file}")
                        all_files.append(found_file)
            self.args['file'] = found_file
            if len(all_files) > 1:
                all_files.sort(key=os.path.getmtime)
                self.args['file'] = all_files[-1]
                util.info(f"Choosing: {self.args['file']} (newest)")
        if not self.args['file']:
            self.error("Couldn't find an XPR Vivado project to open")
        projname = os.path.splitext(os.path.basename(self.args['file']))[0]
        projdir = os.path.dirname(self.args['file'])
        oc_root = util.get_oc_root()
        oc_vivado_tcl = os.path.join(oc_root, 'boards', 'vendors', 'xilinx', 'oc_vivado.tcl')
        command_list = [
            self.vivado_exe, '-source', oc_vivado_tcl,
            '-log', f"{projname}.open.log", self.args['file']
        ]
        self.write_eda_config_and_args()
        self.exec(projdir, command_list)



''' opencos.utils.subprocess_helpers -- wrappers for subprocess to support background/tee'''

import os
import signal
import subprocess
import sys
import threading

import psutil

from opencos import util
from opencos.files import safe_shutil_which
from opencos.util import debug, error, info, warning, progname, global_log
from opencos.utils import status_constants
from opencos.utils.str_helpers import strip_ansi_color

IS_LINUX: bool = sys.platform.startswith('linux')
IS_MACOS: bool = sys.platform.startswith('darwin')
IS_WINDOWS: bool = sys.platform.startswith('win')

# For non-Windows, we track the background parent PIDs, because some tools (vivado XSim,
# most Modelsim/Questa variants) tend to spawn children PIDs that don't always respond
# nicely to a friendly *nix SIGTERM. So we'll remember what our parent PIDs are, and
# eda.py's (or other CLI opencos script) can use signal to cleanup any remaining
# parents + children using subprocess_helpers.cleanup_all()
#
# Note that in non-Windows we run all subprocess with start_new_session=True, to create
# chain them to a PGID group for more effective terminate / killing later. In Windows
# jobs we use: creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
ALL_PARENT_PIDS = set()

def get_command_list(command_list: list, shell: bool = False):
    '''Returns a str if this is non-Windows + shell=True'''
    if not IS_WINDOWS and shell:
        # Return a str:
        return ' '.join(command_list)

    return command_list # leave as list.


def get_kwargs(
        work_dir: str, shell: bool = False, is_subprocess_popen: bool = True
) -> dict:
    '''Returns the subprocess.Popen class ref'''
    proc_kwargs = { 'shell': shell }

    if is_subprocess_popen:
        proc_kwargs.update({
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT
        })

    if work_dir:
        proc_kwargs['cwd'] = work_dir

    bash_exec = safe_shutil_which('bash')
    if shell and bash_exec and not IS_WINDOWS:
        # Note - windows powershell will end up calling: /bin/bash /c, which won't work
        proc_kwargs['executable'] = bash_exec    # for bash features (why we're running shell=True)

    if IS_WINDOWS:
        proc_kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
    else:
        proc_kwargs['start_new_session'] = True

    return proc_kwargs



def subprocess_run(
        work_dir: str, command_list: list, fake: bool = False, shell: bool = False,
        timeout: int = 0
) -> int:
    ''' Run command_list in the foreground, with preference to use bash if shell=True'''

    proc_kwargs = get_kwargs(work_dir=work_dir, shell=shell, is_subprocess_popen=False)

    # For subprocess.run(...) with timeout we simply add to args,
    # pass timeout=value to subprocess.run(..)
    if timeout > 0:
        proc_kwargs['timeout'] = timeout

    c = get_command_list(command_list, shell=shell)

    if fake:
        info(f"subprocess_run FAKE: would have called subprocess.run({c}, **{proc_kwargs}")
        return 0

    debug(f"subprocess_run: About to call subprocess.run({c}, **{proc_kwargs}")
    proc = subprocess.run(c, check=True, **proc_kwargs)
    # Note - we do not get PID management for subprocess_run(...)
    return proc.returncode


# pylint: disable=unknown-option-value
# pylint: disable=too-many-branches,too-many-locals,too-many-statements,too-many-arguments
# pylint: disable=too-many-positional-arguments
def subprocess_run_background(
        work_dir: str, command_list: list, background: bool = True, fake : bool = False,
        shell: bool = False, tee_fpath: str = '',
        timeout: int = 0, timeout_str: str = '',
        progress_timeout: int = 0, progress_timeout_str: str = '',
        force_no_color: bool = False
) -> (str, str, int):
    ''' Run command_list in the background, with preference to use bash if shell=True

    tee_fpath is relative to work_dir.

    Note that stderr is converted to stdout, and stderr is retuned as '':
        Returns tuple of (stdout str, '', int return code)
    '''

    debug(f'subprocess_run_background: {background=} {tee_fpath=} {shell=}')

    if fake:
        # let subprocess_run handle it (won't run anything)
        rc = subprocess_run(work_dir, command_list, fake=fake, shell=shell, timeout=timeout)
        return '', '', rc

    proc_kwargs = get_kwargs(work_dir=work_dir, shell=shell, is_subprocess_popen=True)

    c = get_command_list(command_list, shell=shell)

    debug(f"subprocess_run_background: about to call subprocess.Popen({c}, **{proc_kwargs})")

    # Set-up a watchdog to track timeout, if timeout > 0 (enabled)
    def overall_watchdog_timer_expired(proc) -> None:
        error(f'subprocess_run_background: TimeoutExpired after {timeout} seconds {timeout_str}',
              do_exit=False, error_code=status_constants.EDA_EXEC_SUBPROCESS_TIMEOUT)
        kill_proc_tree(proc.pid)

    # Set-up another watchdog to track progress timeout, if timeout > 0 (enabled)
    def progress_watchdog_timer_expired(proc) -> None:
        error(f'subprocess_run_background: TimeoutExpired after {progress_timeout} seconds',
              f'{progress_timeout_str}',
              do_exit=False, error_code=status_constants.EDA_EXEC_SUBPROCESS_TIMEOUT)
        kill_proc_tree(proc.pid)

    proc = subprocess.Popen(c, **proc_kwargs) # pylint: disable=consider-using-with
    if not background:
        info(f'PID {proc.pid} for {command_list[0]}')
    add_running_parent_pid(proc.pid)

    # This will fire even if readline() is blocked
    if timeout > 0 and proc.pid:
        watchdog = threading.Timer(timeout, overall_watchdog_timer_expired, args=[proc])
        watchdog.start()
    else:
        watchdog = None

    progress_watchdog = None

    def reset_progress_watchdog_timer():
        nonlocal progress_watchdog

        if progress_timeout == 0 or not proc.pid:
            # not enabled, don't start it.
            return

        if progress_watchdog:
            progress_watchdog.cancel()

        # Reuse your existing expire logic
        progress_watchdog = threading.Timer(
            progress_timeout, progress_watchdog_timer_expired,  args=[proc]
        )
        progress_watchdog.start()


    stdout_chunks = []
    tee_fpath_f = None
    reset_progress_watchdog_timer()

    if tee_fpath:
        tee_fpath = os.path.join(work_dir, tee_fpath)
        try:
            tee_fpath_f = open( # pylint: disable=consider-using-with
                tee_fpath, 'w', encoding='utf-8'
            )
        except Exception as e:
            error(f'Unable to open file "{tee_fpath}" for writing, {e}')

    no_color = force_no_color or not util.args['color']
    try:
        for line in iter(proc.stdout.readline, b''):
            line = line.decode("utf-8", errors="replace") # leave \n intact

            reset_progress_watchdog_timer()

            # Since we don't control what the subprocess command did, if it
            # thinks we support color, but user ran with --no-color, we need to strip ANSI colors:
            if no_color:
                line = strip_ansi_color(line)

            # Print the line with color, if --color:
            if not background:
                print(line, end='')

            # for all logs, and the returned stdout str, if we haven't stripped color yet,
            # we need to now, before writing to tee_fpath_f, or to global_log:
            if not no_color:
                line = strip_ansi_color(line)
            line = line.replace('\r', '') # remove CR

            if tee_fpath_f:
                tee_fpath_f.write(line)
            if global_log.file:
                # directly write to file handle, avoid util.UtilLogger.write(line, end='')
                global_log.file.write(line)
            stdout_chunks.append(line)

        # If we broke the for-loop, then stdout has expired and we broke the process.
        proc.communicate()

    finally:
        if watchdog:
            watchdog.cancel()
        if progress_watchdog:
            progress_watchdog.cancel()

    remove_completed_parent_pid(proc.pid)

    rc = proc.returncode
    if tee_fpath_f:
        tee_fpath_f.write(f'INFO: [{progname}] subprocess_run_background: returncode={rc}\n')
        tee_fpath_f.close()
        if not background:
            info('subprocess_run_background: wrote: ' + os.path.abspath(tee_fpath))

    return ''.join(stdout_chunks), '', rc


def add_running_parent_pid(pid: int) -> None:
    '''Adds pid (if still alive) to ALL_PARENT_PIDS'''
    try:
        p = psutil.Process(pid)
        ALL_PARENT_PIDS.add(p.pid)
    except (ProcessLookupError, psutil.NoSuchProcess):
        pass
    except Exception as e:
        error(f'{pid=} exception {e}')

def remove_completed_parent_pid(pid: int) -> None:
    '''Removes pid (if no longer alive) from ALL_PARENT_PIDS.'''
    try:
        p = psutil.Process(pid)
        warning(f'PID {p.pid} still running')
    except (ProcessLookupError, psutil.NoSuchProcess):
        ALL_PARENT_PIDS.remove(pid)
    except Exception as e:
        error(f'{pid=} exception {e}')


def cleanup_all() -> None:
    '''Kills everything from ALL_PARENT_PIDS.'''
    for parent in ALL_PARENT_PIDS:
        kill_proc_tree(parent)


def kill_proc_tree(pid: int) -> None:
    '''
    Kills a process and its entire descendant tree

    Will attempt to use (non-Windows) the process-groupid (pgid)
    '''

    # For Windows:
    #
    # Send CTRL_BREAK to the entire group
    # Note: This requires CREATE_NEW_PROCESS_GROUP at startup
    #os.kill(proc.pid, signal.CTRL_BREAK_EVENT)
    #
    # /F = Force
    # /T = Tree (Kills the process AND all child processes)
    # /PID = The process ID
    #subprocess.run(['taskkill', '/F', '/T', '/PID', str(proc.pid)],
    #               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)


    # 1. Grab PGID immediately if we can. This is to deal with subprocesses
    # shell=True, or closed sourced tools that spawned their own ways:
    pgid = None
    if not IS_WINDOWS:
        try:
            pgid = os.getpgid(pid) # pylint: disable=no-member
        except (ProcessLookupError, OSError):
            pgid = None

    # 2. Get all processes to kill
    all_to_kill = {} # { key=pid: value=psutil.Process }
    try:
        parent = psutil.Process(pid)
        for p in parent.children(recursive=True) + [parent]:
            all_to_kill[p.pid] = p
        info(f'{pid=} {parent=}, {all_to_kill=}')
    except (ProcessLookupError, psutil.NoSuchProcess):
        all_to_kill = {}

    # 2b. Additional fallback, search the system for anyone else in the same
    # PGID to escape processes that above parent.childen(recursive=True) may have
    # missed
    if pgid:
        for p in psutil.process_iter(['pid']):
            if not p:
                continue
            try:
                # Manually fetch the PGID using the standard os module
                # pylint: disable=no-member
                if os.getpgid(p.info['pid']) == pgid and p.info['pid'] not in all_to_kill:
                    all_to_kill[p.info['pid']] = p
            except (ProcessLookupError, psutil.NoSuchProcess, psutil.AccessDenied):
                continue

    # 3. Start gracefull killing (terminate / SIGTERM):
    for p in all_to_kill.values():
        if not p:
            continue
        try:
            info(f'Original {pid=} terminating {p}')
            if IS_WINDOWS:
                # I would like to call:
                #     os.kill(p.pid, signal.CTRL_BREAK_EVENT)
                # But that only works if I have a tty and most Windows python to
                # check isatty() is often unrealiable. We will instead skip to
                # taskkill later
                continue

            # else, non-windows:
            p.terminate() # SIGTERM

        except (ProcessLookupError, psutil.NoSuchProcess):
            pass

    # 4. Wait and ungraceful killing (kill / SIGKILL):
    _, alive = psutil.wait_procs(all_to_kill.values(), timeout=3)
    if alive:
        if IS_WINDOWS:
            # Windows, using subprocess to do the killing:
            # /F = Force, /T = Tree
            subprocess.run( # pylint: disable=subprocess-run-check
                ['taskkill', '/F', '/T', '/PID', str(pid)],
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
        else:
            # non-Windows, target the PGID group to kill escaped children
            if pgid:
                try:
                    info(f'Original {pid=} killing {pgid=}')
                    # pylint: disable=no-member
                    os.killpg(pgid, signal.SIGKILL)
                except Exception:
                    pass

            # non-Windows, final pass on any PIDs psutil still sees:
            for p in alive:
                if not p:
                    continue
                try:
                    info(f'Original {pid=} killing survivor {p}')
                    p.kill()
                except (ProcessLookupError, psutil.NoSuchProcess):
                    pass



'''opencos.utils.dict_helpers -- Various dict helpers functions'''

def dict_diff(d1: dict, d2: dict) -> dict:
    '''Returns dict of added/removed/changed keys in d1 vs d2.

    If an empty dict is returned then there were no differences'''

    diff = {
        'added': {},
        'removed': {},
        'changed': {},
    }

    # Keys added to d2
    for key in d2.keys() - d1.keys():
        diff['added'][key] = d2[key]

    # Keys removed from d1
    for key in d1.keys() - d2.keys():
        diff['removed'][key] = d1[key]

    # Keys present in both
    for key in d1.keys() & d2.keys():
        if d1[key] != d2[key]:
            diff['changed'][key] = f'{d1[key]} --> {d2[key]}'

    for key in ('added', 'removed', 'changed'):
        if not diff[key]:
            del diff[key]

    return diff



''' opencos.tools.slang - Used by opencos.eda for elab commands w/ --tool=slang

Contains classes for ToolSlang, CommandElabSlang
'''

# pylint: disable=R0801 # (calling functions with same arguments)

import os

from opencos import util
from opencos.commands import CommandElab
from opencos.eda_base import Tool
from opencos.files import safe_shutil_which
from opencos.utils import slang_helpers
from opencos.utils.str_helpers import sanitize_defines_for_sh


class ToolSlang(Tool):
    '''ToolSlang used by opencos.eda for --tool=slang'''

    _TOOL = 'slang'
    _EXE = 'slang'
    _URL = 'https://github.com/MikePopoloski/slang'

    slang_exe = ''
    slang_base_path = ''
    slang_tidy_exe = ''
    slang_hier_exe = ''

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        slang_helpers.slang_available.init(exe=self._EXE)
        path = slang_helpers.slang_available.exe
        if not path:
            if self.get_version_allow_certain_command_if_tool_not_installed():
                return self._VERSION
            self.error(f'"{self._EXE}" not in path, need to get it ({self._URL}')
            return ''

        self.slang_exe = path
        self.slang_base_path, _ = os.path.split(path)
        self.slang_tidy_exe = safe_shutil_which(
            os.path.join(self.slang_base_path, 'slang-tidy')
        )
        self.slang_hier_exe = safe_shutil_which(
            os.path.join(self.slang_base_path, 'slang-hier')
        )
        self._VERSION = slang_helpers.slang_available.version
        return self._VERSION

    def set_tool_defines(self):
        super().set_tool_defines()
        if 'SYNTHESIS' not in self.defines:
            self.defines['SIMULATION'] = None # add define
        # Expected to manually add SYNTHESIS command line or target, otherwise.
        # Similarly, you could use --tool slang_yosys for a synthesis friendly
        # elab in Yosys.


class CommandElabSlang(CommandElab, ToolSlang):
    '''CommandElabSlang is a command handler for: eda elab --tool=slang'''

    def __init__(self, config:dict):
        CommandElab.__init__(self, config=config)
        ToolSlang.__init__(self, config=self.config)
        self.args.update({
            'slang-args': [], # aka, --single-unit, --ast-json <fname>, --ast-json-source-info
            'slang-json': False, # sets all the args I know of for AST.
            'slang-top': '',
            'tidy': False, # run slang-tidy instead of slang
            'hier': False, # run slang-hier instead of slang
            'warnings-ok': False
        })

        self.all_json_args = [
            '--ast-json', ## needs filename: slang.json'
            '--ast-json-source-info',
            '--ast-json-detailed-types',
        ]

        self.args_help.update({
            'tidy': "Runs 'slang-tidy' instead of 'slang', with no ast- args.",
            'hier': "Runs 'slang-hier' instead of 'slang', with no ast- args.",
            'warnings-ok': 'Warnings will not fail from eda, they will still be shown',
        })


        # If we're in elab, so not in general ToolSlang, set define for SLANG
        self.defines.update({
            'SLANG': 1
        })

        self.slang_command_lists = []

    def _remove_zero_warnings_from_log_must_strings(self) -> None:
        '''Removes anything about "0 warning" from the log-must-strings'''
        for i, x in enumerate(self.args.get('log-must-strings', [])):
            if '0 warning' in x:
                self.args['log-must-strings'][i] = ''



    # Note that we follow parent class CommandSim's do_it() flow, that way --export args
    # are handled.
    def prepare_compile(self) -> None:
        ''' prepare_compile() - following parent Commandsim's run() flow'''
        self.set_tool_defines()

        if self.args['warnings-ok']:
            self._remove_zero_warnings_from_log_must_strings()

        self.write_eda_config_and_args()

        self.slang_command_lists = self.get_compile_command_lists()
        self.write_slang_sh()

    def compile(self):
        pass

    def elaborate(self):
        ''' elaborate() - following parent CommandSim's run() flow, runs slang_command_lists'''
        if self.args['stop-before-compile'] or \
           self.args['stop-after-compile']:
            return
        self.run_commands_check_logs(self.slang_command_lists)

    def get_compile_command_lists(self, **kwargs) -> list:
        '''Returns list of util.ShellCommandList, for slang we'll run this in elaborate()'''

        command_list = self._get_slang_command_list_start() # slang vs slang-tidy vs slang-hier
        command_list += self.tool_config.get('compile-args', '--single-unit').split()
        command_list += self.args['slang-args'] # add user args.
        command_list += self._get_slang_json_args(command_exe=command_list[0])
        command_list += self._get_slang_tool_config_waivers()

        # incdirs
        for value in self.incdirs:
            command_list += [ '--include-directory', value ]

        # defines:
        if self.args['ext-defines-sv-fname']:
            self.create_ext_defines_sv()
        else:
            for k,v in self.defines.items():
                command_list.append( '--define-macro' )
                if v is None:
                    command_list.append( k )
                else:
                    # Generally we should only support int and str python types passed as
                    # --define-macro {k}={v}
                    command_list.append( f'{k}={sanitize_defines_for_sh(v)}' )

        # parameters
        command_list.extend(
            self.process_parameters_get_list(arg_prefix='-G ')
        )

        # Because many elab target-name won't match the --top needed for
        # slang, we'll leave this to arg --slang-top:
        if self.args.get('slang-top', None):
            command_list += [ '--top', self.args['slang-top'] ]

        # Force -j setting if not already set.
        if not any((a.startswith('-j') or a.startswith('--threads')) for a in command_list):
            command_list += [ '-j', str(self.args['threads']) ]

        command_list += self.files_sv + self.files_v

        command_list = util.ShellCommandList(command_list, tee_fpath='compile.log')
        return [command_list]

    def write_slang_sh(self):
        '''Writes run_slang.sh to work-dir'''
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='run_slang.sh',
            command_lists=self.slang_command_lists,
            line_breaks=True
        )

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # all handled by get_compile_command_lists()
        return []

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []

    def update_tool_warn_err_counts_from_log_lines(
            self, log_lines: list, bad_strings: list, warning_strings: list
    ) -> None:
        '''
        Overriden from Command, we ignore bad_strings/warning_strings and use a custom
        checker.
        '''
        for line in log_lines:
            if not line.startswith('Build failed: ') and \
               not line.startswith('Build succeeded: '):
                continue
            if not all(x in line for x in ('errors', 'warnings')):
                continue

            parts = line.strip().split()
            if len(parts) < 6:
                continue

            errs = parts[2]
            warns = parts[4]
            if errs.isdigit():
                self.tool_error_count += int(errs)
            if warns.isdigit():
                self.tool_warning_count += int(warns)

    def _get_slang_command_list_start(self) -> list:
        command_list = [self.slang_exe]

        if self.args['tidy']:
            if not safe_shutil_which(self.slang_tidy_exe):
                util.warning("Running tool slang with --tidy, but 'slang-tidy'",
                             "not in PATH, using 'slang' instead")
            else:
                command_list = [self.slang_tidy_exe]

        if self.args['hier']:
            if self.args['tidy']:
                util.warning('Running with --tidy and --heir, will attempt to use slang-hier')
            elif not safe_shutil_which(self.slang_hier_exe):
                util.warning("Running tool slang with --hier, but 'slang-hier'",
                             "not in PATH, using 'slang' instead")
            else:
                command_list = [self.slang_hier_exe]

        return command_list

    def _get_slang_json_args(self, command_exe: str) -> list:
        command_list = []

        _, command_exe_leaf = os.path.split(command_exe)
        if self.args.get('slang-json', False) and command_exe_leaf == 'slang':
            for arg in self.all_json_args:
                if arg not in command_list:
                    command_list.append(arg)
                    if arg == '--ast-json': # needs filename
                        command_list.append('slang.json')
                        util.artifacts.add(
                            name=os.path.join(self.args['work-dir'], 'slang.json'),
                            typ='json',
                            description='Abstract syntax tree from slang --ast-json'
                        )

        return command_list

    def _get_slang_tool_config_waivers(self) -> list:
        # Add compile waivers from config and command-line args
        return [f'-Wno-{waiver}' for waiver in
                self.tool_config.get('compile-waivers', []) + self.args['compile-waivers']]


class CommandLintSlang(CommandElabSlang):
    '''CommandLintSlang is a command handler for: eda lint --tool=slang.'''

    command_name = 'lint'

    def __init__(self, config: dict):
        super().__init__(config)
        # keep stop-after-compile=False, allow's CommandElabSlang.elaborate() to run.
        self.args['slang-args'] = ['--lint-only', '--ignore-unknown-modules']



''' opencos.tools.yosys - base class for slang_yosys.py, invio_yosys.py, tabbycad_yosys.py

Contains classes for ToolYosys
'''

# pylint: disable=R0801 # (calling functions with same arguments)

import os
import subprocess

from opencos import util, eda_config
from opencos.commands import CommandSynth, CommandLec, CommandFList
from opencos.eda_base import Tool, get_eda_exec
from opencos.files import safe_shutil_which
from opencos.utils.markup_helpers import yaml_safe_load

from opencos.commands.sim import parameters_dict_get_command_list


def get_commands_to_run_scriptfiles(
        script_fnames_list: list, yosys_exe: str
) -> [util.ShellCommandList]:
    '''Checks file existence and returns list of commands to run a

    list of yoysys script(s)'''

    if not script_fnames_list:
        return []

    yosys_cmdlists = []
    for i,fpath in enumerate(script_fnames_list):
        if not os.path.isfile(fpath):
            util.error(f'yosys-scriptfile={fpath} file does not exist')
        cmdlist = util.ShellCommandList(
            [yosys_exe, '--scriptfile', os.path.abspath(fpath)],
            tee_fpath = f'yosys_scriptfile.{i}.log'
        )
        yosys_cmdlists.append(cmdlist)
    return yosys_cmdlists


class ToolYosys(Tool):
    '''Parent class for ToolTabbyCadYosys, ToolInvioYosys, ToolSlangYosys'''

    _TOOL = 'yosys'
    _EXE = 'yosys'
    _URL = 'https://yosyshq.readthedocs.io/en/latest/'

    yosys_exe = ''
    sta_exe = ''
    sta_version = ''

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        path = safe_shutil_which(self._EXE)
        if not path:
            if self.get_version_allow_certain_command_if_tool_not_installed():
                return self._VERSION
            self.error(f'"{self._EXE}" not in path or not installed, see {self._URL}')
        else:
            self.yosys_exe = path

        # Unforunately we don't have a non-PATH friendly support on self._EXE to set
        # where standalone 'sta' is. Even though Yosys has 'sta' internally, Yosys does
        # not fully support timing constraints or .sdc files, so we have to run 'sta'
        # standalone.
        sta_path = safe_shutil_which('sta')
        if sta_path:
            util.debug(f'Also located "sta" via {sta_path}')
            self.sta_exe = sta_path
            sta_version_ret = subprocess.run(
                [self.sta_exe, '-version'], capture_output=True, check=False
            )
            util.debug(f'{self.yosys_exe} {sta_version_ret=}')
            sta_ver = sta_version_ret.stdout.decode('utf-8', errors='replace').split()
            if sta_ver and isinstance(sta_ver, list):
                self.sta_version = sta_ver[0]

        version_ret = subprocess.run(
            [self.yosys_exe, '--version'], capture_output=True, check=False
        )
        util.debug(f'{self.yosys_exe} {version_ret=}')

        # Yosys 0.48 (git sha1 aaa534749, clang++ 14.0.0-1ubuntu1.1 -fPIC -O3)
        words = version_ret.stdout.decode('utf-8', errors='replace').split()

        if len(words) < 2:
            util.warning(f'{self.yosys_exe} --version: returned unexpected str {version_ret=}')
        self._VERSION = words[1]
        return self._VERSION

    def set_tool_defines(self):
        super().set_tool_defines()
        if 'OC_LIBRARY' not in self.defines:
            self.defines.update({
                'OC_LIBRARY_BEHAVIORAL': None,
                'OC_LIBRARY': "0"
            })

    # What is supported by get_yosys_defines_incdirs_files():
    flist_yosys_command_choices = [
        'read_verilog',
        'read_slang',
    ]

    def get_yosys_defines_incdirs_files( # pylint: disable=too-many-branches
            self,
            for_synthesis: bool = True,
            yosys_command: str = 'read_verilog',
            **kwargs,
    ) -> list:
        '''
        Returns a partial list of all the args for a read_verilog or read_slang command in yosys
        Handles defines, incdirs, files_sv, files_v

        self, in this case, is intended to be a ToolYosys AND opencos.eda_base.CommandDesign derived
        class object.

        This should be used by CommandFList or CommandSynth derivations for yosys tools:
          *Yosys, *SlangYosys

        Valid yosys_command(s) are self.flist_yosys_command_choices

        Valid kwargs bools (default True if ommitted) - defines, incdirs, verilog, systemverilog,
          top
        '''

        ret_list = []

        defines = getattr(self, 'defines', {})
        incdirs = getattr(self, 'incdirs', [])
        files_v = getattr(self, 'files_v', [])
        files_sv = getattr(self, 'files_sv', [])
        top  = getattr(self, 'args', {}).get('top', '')

        if yosys_command not in self.flist_yosys_command_choices:
            self.error(
                f'Unsupported {yosys_command=} arg on call to',
                f'get_yosys_defines_incdirs_files({for_synthesis=}, {yosys_command=}, {kwargs}),',
                'run with --debug for stacktrace'
            )

        if kwargs.get('defines', True):
            for name, value in defines.items():
                if not name:
                    continue
                if for_synthesis and name in ['SIMULATION']:
                    continue

                if yosys_command == 'read_slang':
                    if value is None:
                        ret_list.append(f'--define-macro {name}')
                    else:
                        ret_list.append(f'--define-macro {name}={value}')
                else:
                    # assume read_verilog
                    if value is None:
                        ret_list.append(f'-D{name}')
                    else:
                        ret_list.append(f'-D{name}={value}')

            # We must define SYNTHESIS for oclib_defines.vh to work correctly.
            if for_synthesis and 'SYNTHESIS' not in defines:
                if yosys_command == 'read_slang':
                    ret_list.append('--define-macro SYNTHESIS')
                else:
                    ret_list.append('-DSYNTHESIS')

        if kwargs.get('incdirs', True):
            for path in incdirs:
                if yosys_command == 'read_slang':
                    ret_list.append(f'-I {path}')
                else:
                    # assume read_verilog
                    ret_list.append(f'-I{path}')
        if kwargs.get('verilog', True):
            ret_list.extend(files_v)
        if kwargs.get('systemverilog', True):
            ret_list.extend(files_sv)
        if yosys_command != 'read_verilog' and kwargs.get('top', True) and top:
            # read_verilog has no '--top' option.
            ret_list.append(f'--top {top}')

        return ret_list


class CommonSynthYosys(CommandSynth, ToolYosys):
    '''Common parent class used by invio_yosys and tabbycad_yosys

    for child classes: CommandSynthInvioYosys and tabbycad_yosys.CommandSynthTabbyCadYosys
    '''

    def __init__(self, config: dict):
        CommandSynth.__init__(self, config=config)
        ToolYosys.__init__(self, config=self.config)

        self.args.update({
            'sta': False,
            'liberty-file': '',
            'sdc-file': '',
            'yosys-synth': 'synth',              # synth_xilinx, synth_altera, etc (see: yosys help)
            'yosys-pre-synth': ['prep', 'proc'], # command run in yosys prior to yosys-synth.
            'yosys-blackbox': [],                # list of modules that yosys will blackbox.
            'yosys-scriptfile': [],
            'sta-scriptfile': [],
            'rename-module': ''
        })
        self.args_help.update({
            'sta': (
                'After running Yosys, run "sta" with --liberty-file.'
                ' sta can be installed via: https://github.com/The-OpenROAD-Project/OpenSTA'
            ),
            'sdc-file': (
                '.sdc file to use with --sta, if not present will use auto constraints.'
                ' Note you can have .sdc files in "deps" of DEPS.yml targets.'
            ),
            'liberty-file': (
                'Single liberty file for synthesis and sta,'
                ' for example: github/OpenSTA/examples/nangate45_slow.lib.gz'
            ),
            'yosys-synth': 'The synth command provided to Yosys, see: yosys help.',
            'yosys-pre-synth': (
                'Yosys commands performed prior to running "synth"'
                ' (or eda arg value for --yosys-synth)'
            ),
            'yosys-blackbox': (
                'List of modules that yosys will blackbox, likely will need these'
                ' in Verilog-2001 for yosys to read outside of slang and synth'
            ),
            'yosys-scriptfile': (
                'Instead of using a built-in flow from eda, use your own scripts that are called'
                ' via: yosys --scriptfile <this-arg>. You can set multiple args for multiple'
                ' scriptfile (appends). NOTE: the default for eda is to run in a work-dir such as'
                ' ./eda.work/<target>.synth/, which means you may want to use absolute paths for'
                ' source files (.sv, etc) in your custom scriptfile, or use ../.. to get back to '
                ' the current directory. If you want to run in place and avoid'
                ' ./eda.work/<target>.synth/ you can use arg --work-dir=. or to run in place from'
                ' the DEPS target directory, use --work-dir-use-target-dir'
            ),
            'sta-scriptfile': (
                'Instead of using a built-in flow from eda, use your own script that is called'
                ' via: sta -no_init -exit <this-arg>.  You can set multiple args for multiple'
                ' scriptfile (appends)'
            ),
            'rename-module': 'Renames the output .v and module name',
        })

        self.yosys_out_dir = ''
        self.yosys_v_path = ''
        self.full_work_dir = ''
        self.blackbox_list = []
        self.top_module = ''

    def do_it(self) -> None:
        self.set_tool_defines()
        self.write_eda_config_and_args()

        # Set up some dirs and filenames.
        self.full_work_dir = self.args.get('work-dir', '')
        if not self.full_work_dir:
            self.error(f'work_dir={self.full_work_dir} is not set')
        self.full_work_dir = os.path.abspath(self.full_work_dir)
        self.yosys_out_dir = os.path.join(self.full_work_dir, 'yosys')
        util.safe_mkdir(self.yosys_out_dir)
        self.yosys_v_path = os.path.join(self.yosys_out_dir, f'{self.args["top"]}.v')

        if self.is_export_enabled():
            self.do_export()
            return

        if self.args['yosys-scriptfile']:
            yosys_cmdlists = self.get_commands_user_yosys_scriptfile()
            sta_cmdlists = self.create_sta_f() # works for --sta w/out BYO scripts.

            # We create a run_yosys.sh wrapping these scripts, but we do not run this one.
            util.write_shell_command_file(
                dirpath=self.args['work-dir'],
                filename='run_yosys.sh',
                command_lists=(yosys_cmdlists + sta_cmdlists)
            )

            # actually run it.
            for x in yosys_cmdlists + sta_cmdlists:
                if x:
                    self.exec(work_dir=self.full_work_dir, command_list=x,
                              tee_fpath=x.tee_fpath)

        else:
            self.write_and_run_yosys_f_files()


    def get_commands_user_yosys_scriptfile(self) -> [util.ShellCommandList]:
        '''Checks file existence and returns list of commands to run a

        list of yoysys script(s)'''
        cmd_lists = get_commands_to_run_scriptfiles(
            script_fnames_list=self.args['yosys-scriptfile'],
            yosys_exe=self.yosys_exe
        )

        if not cmd_lists:
            util.error('Could not generate yosys commands for scripts',
                       f'{self.args["yosys-scriptfile"]}')

        return cmd_lists


    def get_commands_user_sta_scriptfile(self) -> [util.ShellCommandList]:
        '''Checks file existence and returns list of commands'''
        if not self.args['sta-scriptfile']:
            return []

        # Add URL info for OpenSTA source code:
        self._add_opensta_info()

        ret_list = []
        for i,fpath in enumerate(self.args['sta-scriptfile']):
            if not os.path.isfile(fpath):
                self.error(f'sta-scriptfile={fpath} file does not exist')
            cmdlist = util.ShellCommandList(
                [self.sta_exe, '-no_init', '-exit', os.path.abspath(fpath)],
                tee_fpath = f'sta_scriptfile.{i}.log'
            )
            ret_list.append(cmdlist)
        return ret_list


    def write_and_run_yosys_f_files(self) -> None:
        '''Derived classes may override, to run remainder of do_it() steps

        These built-ins do not use slang or another SV preprocessing step.
        1. Creates and runs: yosys.synth.f
           -- does blackboxing and synth steps
        4. Creates a wrapper for human debug and reuse: yosys.f
        '''

        # Note - big assumption here that "module myname" is contained in myname.[v|sv]:
        # we use both synth-blackbox and yosys-blackbox lists to blackbox modules in the
        # yosys step (not in the slang step)
        self.blackbox_list = self.args.get('yosys-blackbox', [])
        self.blackbox_list += self.args.get('synth-blackbox', [])

        # work-dir / yosys has already been created.

        # Create and run yosys.synth.f
        synth_command_list = self.create_yosys_synth_f() # util.ShellCommandList

        # Optinally create and run a sta.f:
        sta_command_lists = self.create_sta_f() # [] or [util.ShellCommandList]

        # We create a run_yosys.sh wrapping these scripts, but we do not run this one.
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='run_yosys.sh',
            command_lists=[synth_command_list] + sta_command_lists,
        )

        # Do not run this if args['stop-before-compile'] is True
        if self.args.get('stop-before-compile', False):
            return

        # Run the synth commands standalone:
        self.exec(work_dir=self.full_work_dir, command_list=synth_command_list,
                  tee_fpath=synth_command_list.tee_fpath)

        for x in sta_command_lists:
            if self.args['sta'] and x:
                self.exec(work_dir=self.full_work_dir, command_list=x,
                          tee_fpath=x.tee_fpath)

        if self.status == 0:
            util.info(f'yosys: wrote verilog to {self.yosys_v_path}')


    def create_yosys_synth_f(self) -> util.ShellCommandList:
        '''Derived classes may define, if they wish to get a list of yosys commands'''

        # Create yosys.synth.f
        yosys_synth_f_path = os.path.join(self.full_work_dir, 'yosys.synth.f')

        # Since this assumes we didnt' run a SystemVerilog pre-processing step,
        # read in all the verilog
        yosys_blackbox_list = self.get_yosys_blackbox_list()

        if self.args['liberty-file'] and not os.path.exists(self.args['liberty-file']):
            self.error(f'--liberty-file={self.args["liberty-file"]} file does not exist')

        with open(yosys_synth_f_path, 'w', encoding='utf-8') as f:
            lines = [
                self._get_read_verilog_one_liner()
            ]

            if self.args['liberty-file']:
                lines.append('read_liberty -lib ' + self.args['liberty-file'])

            for inst in yosys_blackbox_list:
                lines.append('blackbox ' + inst)

            lines += self.get_synth_command_lines()
            f.write('\n'.join(lines))

        synth_command_list = util.ShellCommandList(
            [self.yosys_exe, '--scriptfile', 'yosys.synth.f'],
            tee_fpath = 'yosys.synth.log'
        )
        return synth_command_list


    def get_synth_command_lines(self) -> list:
        '''Common yosys tcl after all blackbox and read_verilog commands'''

        lines = []
        lines += self.args.get('yosys-pre-synth', [])

        synth_command = self.args.get('yosys-synth', 'synth')
        if self.args['flatten-all']:
            synth_command += ' -flatten'

        lines.append(synth_command)


        # TODO(drew): I need a blackbox flow here? Or a memory_libmap?
        #   --> https://yosyshq.readthedocs.io/projects/yosys/en/latest/cmd/memory_libmap.html
        # TODO(drew): can I run multiple liberty files?
        if self.args['liberty-file']:
            lines += [
                'dfflibmap -liberty ' + self.args['liberty-file'],
                #'memory_libmap -lib ' + self.args['liberty-file'], # Has to be unzipped?
                'abc -liberty  ' + self.args['liberty-file'],
            ]
        lines += [
            'opt_clean',
        ]
        if self.args['rename-module']:
            lines += [f'rename {self.args["top"]} {self.args["rename-module"]}']
            self.top_module = self.args['rename-module']
        else:
            self.top_module = self.args["top"]
        lines += [
            f'write_verilog {self.yosys_v_path}',
            f'write_json {self.yosys_v_path}.json',
        ]
        return lines

    def get_yosys_blackbox_list(self) -> list:
        '''Returns blackbox list, since we don't have a preprocessing step like

        slang, simply return self.blackbox_list. Intended to be overwritten by
        derived classes so they can blackbox post-preprocessing.
        '''
        return self.blackbox_list


    def create_sta_f(self) -> [util.ShellCommandList]:
        '''Returns command list, for running 'sta' on sta.f'''

        if not self.args['sta']:
            return []

        if self.args['sta-scriptfile']:
            # User brought one or more scriptfiles for STA, use those.
            return self.get_commands_user_sta_scriptfile()

        if not self.args['liberty-file']:
            self.error('--sta is set, but need to also set --liberty-file=<file>')

        if self.args['sdc-file']:
            if not os.path.exists(self.args['sdc-file']):
                self.error(f'--sdc-file={self.args["sdc-file"]} file does not exist')

        if not self.sta_exe:
            self.error(f'--sta is set, but "sta" was not found in PATH, see: {self._URL}')

        # Add URL info for OpenSTA source code:
        self._add_opensta_info()

        sta_command_list = util.ShellCommandList(
            [ self.sta_exe, '-no_init', '-exit', 'sta.f' ],
            tee_fpath = 'sta.log'
        )

        # Need to create sta.f:
        if self.args['sdc-file']:
            sdc_path = self.args['sdc-file']
        elif self.files_sdc:
            # Use files from DEPS target or command line.
            sdc_path = ''
        else:
            # Need to create sdc.f:
            sdc_path = 'sdc.f'
            self.create_sdc_f()

        with open(os.path.join(self.args['work-dir'], 'sta.f'), 'w',
                  encoding='utf-8') as f:
            lines = [
                'read_liberty ' + self.args['liberty-file'],
                'read_verilog ' + self.yosys_v_path,
                'link_design ' + self.top_module,
            ]
            for _file in self.files_sdc:
                lines.append('read_sdc ' + _file)
            if sdc_path:
                lines.append('read_sdc ' + sdc_path)

            lines.append('report_checks')

            f.write('\n'.join(lines))

        # return list with our one generated command-list
        return [util.ShellCommandList(
            sta_command_list,
            tee_fpath = sta_command_list.tee_fpath
        )]


    def create_sdc_f(self) -> None:
        '''Returns None, creates sdc.f'''

        if self.args['sdc-file']:
            # already exists from args, return b/c nothing to create.
            return

        with open(os.path.join(self.args['work-dir'], 'sdc.f'), 'w',
                  encoding='utf-8') as f:
            clock_name = self.args['clock-name']
            period = self.args['clock-ns']
            name_not_equal_clocks_str = f'NAME !~ "{clock_name}"'
            lines = [
                f'create_clock -add -name {clock_name} -period {period} [get_ports ' \
                + '{' + clock_name + '}];',
                f'set_input_delay -max {self.args["idelay-ns"]} -clock {clock_name}' \
                + ' [get_ports * -filter {DIRECTION == IN && ' \
                + name_not_equal_clocks_str + '}];',
                f'set_output_delay -max {self.args["odelay-ns"]} -clock {clock_name}' \
                + ' [get_ports * -filter {DIRECTION == OUT}];',
            ]
            f.write('\n'.join(lines))

    def _add_opensta_info(self) -> None:
        '''Adds OpenSTA URL information to artifacts and logs'''

        opensta_repo_url = 'https://github.com/The-OpenROAD-Project/OpenSTA'

        # Log GPL3.0 license information
        util.info(f'Using OpenSTA (see URL for license and source): {opensta_repo_url}')
        if hasattr(self, 'sta_version') and self.sta_version:
            util.info(f'OpenSTA version: {self.sta_version}')

    def _get_read_verilog_one_liner(self) -> str:
        '''Returns a string, intended to be used w/out Slang, for Verilog or simple

        SV designs'''

        read_verilog_cmd = [
            'read_verilog',
            '-sv',
            '-icells',
        ]
        read_verilog_cmd += self.get_yosys_defines_incdirs_files(
            for_synthesis=True, yosys_command='read_verilog'
        )
        return ' '.join(read_verilog_cmd)


class CommandLecYosys(CommandLec, ToolYosys):
    '''Command handler for: eda lec --designs=<target1> --designs=<target2> --tool=yosys

    Also supports: eda lec --tool=yosys <target>
    If the target sets two args for --designs
    '''

    def __init__(self, config: dict):
        CommandLec.__init__(self, config=config)
        ToolYosys.__init__(self, config=self.config)

        self.args.update({
            'yosys-scriptfile': [],
            'pre-read-verilog': [],
        })
        self.args_help.update({
            'yosys-scriptfile': (
                'Instead of using a built-in flow from eda, use your own scripts that are called'
                ' via: yosys --scriptfile <this-arg>. You can set multiple args for multiple'
                ' scriptfile (appends)'
            ),
            'pre-read-verilog': 'Additional verilog files to read prior to running LEC',
        })

        self.synth_work_dirs = [
            os.path.join('eda.work', 'lec.Design1.synth'),
            os.path.join('eda.work', 'lec.Design2.synth')
        ]

        self.synth_designs_tops = [None, None]
        self.synth_designs_fpaths = [None, None]

    def get_synth_result_fpath(self, target: str) -> str:
        '''Overridden from CommandLec'''

        # Read the eda_output_config.yml, find the "top", and find the output .v filename.
        return ""

    def get_synth_command_list(self, design_num: int) -> list:
        '''Returns one of the synthesis command lists, for design_num=0 or 1'''

        if not design_num in (0, 1):
            self.error(f'{design_num=} we only support LEC on designs 0 and 1')

        synth_cmd_list = [
            get_eda_exec('synth'),
            'synth',
        ]

        if self.args['tool']:
            synth_cmd_list.append('--tool=' + self.args['tool'])

        synth_cmd_list += [
            '--work-dir=' + self.synth_work_dirs[design_num],
            self.args['designs'][design_num],
            f'--rename-module=Design{design_num + 1}',
        ]

        if self.args['flatten-all']:
            # We have to do this or may get conflicts on black-boxed modules, but can
            # be avoided with --no-flatten-all.
            synth_cmd_list.append(
                '--flatten-all'
            )

        self.synth_designs_tops[design_num] = f'Design{design_num + 1}'

        return synth_cmd_list


    def get_synth_top_from_output_config(self, design_num: int) -> (str, str):
        '''Returns the (orignal top name, module name) tuple given the design number

        that we synthesized'''

        work_dir = self.synth_work_dirs[design_num]
        output_cfg_fpath = os.path.join(work_dir, eda_config.EDA_OUTPUT_CONFIG_FNAME)
        data = yaml_safe_load(output_cfg_fpath)
        top = data.get('args', {}).get('top', '')
        rename_module = data.get('args', {}).get('rename-module', '')
        if not top and not rename_module:
            self.error(f'"top" not found in synth run from {work_dir=} in',
                       f'config {output_cfg_fpath}')
        if not rename_module:
            return top, top
        return top, rename_module


    def get_synth_results_fpath(self, design_num: int, top: str) -> str:
        '''Returns the synthesized .v file fpath, using orignal top (not renamed)'''
        if not top:
            top, _ = self.get_synth_top_from_output_config(design_num=design_num)

        work_dir = self.synth_work_dirs[design_num]
        fpath = os.path.join(work_dir, 'yosys', f'{top}.v')
        if not os.path.isfile(fpath):
            self.error(f'{fpath=} does not exists, looking for synth results for LEC {design_num=}')
        return fpath


    def do_it(self) -> None: # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        self.set_tool_defines()
        self.write_eda_config_and_args()

        pwd = os.getcwd()

        if not self.args['top']:
            self.args['top'] = 'yosys_lec'

        if self.args['yosys-scriptfile']:
            yosys_cmdlists = get_commands_to_run_scriptfiles(
                script_fnames_list=self.args['yosys-scriptfile'],
                yosys_exe=self.yosys_exe
            )

            # We create a run_yosys.sh wrapping these scripts, but we do not run this one.
            util.write_shell_command_file(
                dirpath=self.args['work-dir'],
                filename='run_yosys.sh',
                command_lists=yosys_cmdlists
            )

            # actually run it.
            for x in yosys_cmdlists:
                if x:
                    self.exec(work_dir=self.args['work-dir'], command_list=x,
                              tee_fpath=x.tee_fpath)

            util.info(f'LEC ran via --yosys-scriptfile: {self.args["yosys-scriptfile"]}')
            return


        if self.args['synth']:
            synth1_cmd_list = self.get_synth_command_list(design_num=0)
            synth2_cmd_list = self.get_synth_command_list(design_num=1)

            util.info(f'LEC {synth1_cmd_list=}')
            util.info(f'LEC {synth2_cmd_list=}')

            self.exec(pwd, synth1_cmd_list, background=True)
            util.info(f'Finished with 1st LEC synthesis {self.args["designs"][0]}')

            self.exec(pwd, synth2_cmd_list, background=True)
            util.info(f'Finished with 2nd LEC synthesis {self.args["designs"][1]}')


            top0, module0 = self.get_synth_top_from_output_config(design_num=0)
            top1, module1 = self.get_synth_top_from_output_config(design_num=1)
            self.synth_designs_tops = [module0, module1]
            util.info(f'Design tops: {self.synth_designs_tops}')

            # read the output config
            self.synth_designs_fpaths = [
                os.path.abspath(
                    self.get_synth_results_fpath(design_num=0, top=top0)),
                os.path.abspath(
                    self.get_synth_results_fpath(design_num=1, top=top1))
            ]
            util.info(f'Design tops: {self.synth_designs_fpaths}')

        else:
            # don't run synthesis, need the two top level .v|.sv files in
            # self.synth_designs_fpaths, and need the two top module names in
            # self.synth_designs_tops
            self.synth_designs_fpaths = [
                os.path.abspath(self.args['designs'][0]),
                os.path.abspath(self.args['designs'][1])
            ]

            for i in (0, 1):
                if not os.path.isfile(self.args['designs'][i]):
                    self.error(
                        'Using synth=False (--no-synth) --designs=<value> must be a single',
                        f'filename, however {self.args["designs"][i]} does not exist'
                    )

            path, fname = os.path.split(self.synth_designs_fpaths[0])
            module_guess, _ = os.path.splitext(fname)
            top1 = util.get_inferred_top_module_name(
                module_guess=module_guess, module_fpath=self.synth_designs_fpaths[0]
            )
            util.debug(f'design1 {module_guess=} {fname=} {path=}')
            util.info(f'design1 top module name = {top1} (from {path} / {fname})')

            path, fname = os.path.split(self.synth_designs_fpaths[1])
            module_guess, _ = os.path.splitext(fname)
            top2 = util.get_inferred_top_module_name(
                module_guess=module_guess, module_fpath=self.synth_designs_fpaths[1]
            )
            util.debug(f'design2 {module_guess=} {fname=} {path=}')
            util.info(f'design2 top module name = {top2} (from {path} / {fname})')

            self.synth_designs_tops = [top1, top2]

        # Need to create final LEC yosys script, that reads our two designs and runs
        # LEC. Note the designs must have different module names
        if self.synth_designs_tops[0] == self.synth_designs_tops[1]:
            self.error('Cannot run Yosys LEC on two designs with the same top module name:',
                       f'{self.synth_designs_tops}')

        lec_cmd_f_list = []
        if self.args['pre-read-verilog']:
            for x in self.args['pre-read-verilog']:
                if os.path.isfile(x):
                    lec_cmd_f_list += [
                        'read_verilog -sv -icells ' + os.path.abspath(x)
                    ]
                else:
                    self.error(f' --pre-read-verilog file {x} does not exist')

        nooverwrite = ''
        if self.args['synth'] and not self.args['flatten-all']:
            # If we don't flatten-all from synthesis, and we had to run synthesis,
            # then read the 2nd file with -overwrite
            nooverwrite = '-nooverwrite'

        lec_cmd_f_list += [
            '# Design1 (module):',
            f'read_verilog -sv -icells {self.synth_designs_fpaths[0]}',
            '# Design2 (module):',
            f'read_verilog -sv -icells {self.synth_designs_fpaths[1]} {nooverwrite}',
            'clk2fflogic;',
            f'miter -equiv -flatten {" ".join(self.synth_designs_tops)} miter',
            ('sat -seq 50 -verify -prove trigger 0 -show-all -show-inputs -show-outputs'
             ' -set-init-zero miter'),
        ]

        lec_cmd_f_fpath = os.path.join(self.args['work-dir'], 'yosys_lec.f')
        with open(lec_cmd_f_fpath, 'w', encoding='utf-8') as f:
            f.write('\n'.join(lec_cmd_f_list) + '\n')

        lec_cmd_list = 'yosys --scriptfile yosys_lec.f'.split()
        util.info(f'LEC running {lec_cmd_list}')
        self.exec(self.args['work-dir'], lec_cmd_list)


class CommonFListYosys(CommandFList, ToolYosys):
    '''
    Handler for: eda flist --tool=yosys

    This will create a eda.yosys_read_verilog.f flist (or to std-out if --print-to-stdout),
    suitable for args passed to a "read_verilog" command in yosys, such as:

        read_verilog -f FILE_CREATED_BY_THIS_HANDLER [other args]

    '''

    def __init__(self, config: dict, **kwargs):
        CommandFList.__init__(self, config=config,
                              # Added bonus - remove some CommandFList args that we don't
                              # need:
                              include_prefix_args=False, include_quote_args=False,
                              include_format_define_args=False)
        ToolYosys.__init__(self, config=self.config)

        self.args.update({
            'force': True, # always overwrite the output
            'out': 'eda_flist.yosys.read_verilog.f',

            'emit-define'        : True,
            'emit-parameter'     : True,
            'emit-incdir'        : True,
            'emit-plusargs'      : True,
            'emit-v'             : True,
            'emit-sv'            : True,
            'emit-vhd'           : True,
            'emit-cpp'           : False,
            'emit-non-sources'   : False, # as comments, from DEPS 'reqs'
            'flist-yosys-command' : 'read_verilog',
        })
        self.args_help.update({
            'flist-yosys-command': 'The Yosys Command that the genarted dot-f file supports'
        })
        self.args_kwargs.update({
            'flist-yosys-command': { 'choices': list(self.flist_yosys_command_choices) }
        })

    def get_flist_plusargs_list(self) -> list:
        '''Overriden from CommandFList.'''
        if self.args['unprocessed-plusargs']:
            util.warning(f'Command "flist" for --tool={self._TOOL} is not intended for simulation',
                         'and plusargs were present. They will NOT be included in the flist:',
                         f'{self.args["unprocessed-plusargs"]}')

        return []


    def get_write_flist_lines(
            self, add_comment_lines: bool = False
    ) -> list:
        '''Override from CommandFList'''

        # Want: defines + parameters + incdirs + files + --top=VALUE, similar to
        #    how we call ToolYosys.get_yosys_defines_incdirs_files(),
        #    but without the additional args

        ret = []

        # Use our overriden method to post warnings about brining simulation Plusargs to
        # a Yosys tool:
        _ = self.get_flist_plusargs_list()

        if add_comment_lines:
            ret.append(f"## {self.args=}")

            if self.files_non_source:
                ret.append('## reqs (non-source files that are dependencies):')
                for f in self.files_non_source:
                    if self.args['emit-rel-path']:
                        f = os.path.relpath(f)
                    ret.append(f'##   {f}')

        # Note - top is included in this:
        include_items = []
        if self.args['emit-define']:
            include_items.append('defines')
        if self.args['emit-incdir']:
            include_items.append('incdirs')

        ret.extend(self.get_yosys_defines_incdirs_files(
            for_synthesis=True,
            yosys_command=self.args['flist-yosys-command'],
            # True/False kwargs if you want that emitted:
            defines=self.args['emit-define'],
            incdirs=self.args['emit-incdir'],
            verilog=self.args['emit-v'],
            systemverilog=self.args['emit-sv'],
            top=True
        ))

        # and parameters:
        ret.extend(parameters_dict_get_command_list(
            params=self.parameters, arg_prefix='-G '
        ))

        return ret



'''Because so many tools use the exe vsim, I don't want to run `vsim -version` N times for N
tools each to figure out if it's full-Questa, Modelsim, QuestaFSE, or Riviera.

Instead, eda.py can call this once, and then query if this tool exists when running
opencos.eda.auto_tool_setup(..)
'''

import os
import subprocess

from opencos.files import safe_shutil_which
from opencos.util import debug

INIT_HAS_RUN = False  # pylint: disable=invalid-name
TOOL_PATH = {
    'riviera': None,
    'modelsim_ase': None,
    'questa' : None,
    'questa_fe' : None,
    'questa_fse': None
}
TOOL_VERSION = {
}

def lookup_exe(tool: str) -> str:
    '''Fancy get method for the tool path'''
    if not INIT_HAS_RUN:
        init()
    p = TOOL_PATH.get(tool, '')
    if isinstance(p, str):
        return p
    return '' # don't return None

def lookup_version(tool: str) -> str:
    '''Fancy get method for the tool version'''
    if not INIT_HAS_RUN:
        init()
    v = TOOL_VERSION.get(tool, '')
    if isinstance(v, str):
        return v
    return '' # don't return None

def found() -> str:
    '''Returns the found tool exe PATH, or blank str'''
    if not INIT_HAS_RUN:
        init()
    for k,v in TOOL_PATH.items():
        if k and v:
            return k
    return ''

def init_get_version(dirpath: str) -> None: # pylint: disable=too-many-branches
    '''Runs vsim -version using vsim_path, updates TOOL_PATH'''

    vsim_path = safe_shutil_which(os.path.join(dirpath, 'vsim'))
    if not os.access(vsim_path, os.X_OK):
        debug(f'{vsim_path} is not executable')
        vsim_path = ''

    if not vsim_path:
        return # didn't find a valid vsim exe in dirpath (Windows names included)

    proc = None
    try:
        proc = subprocess.run([vsim_path, '-version'], capture_output=True, check=False)
    except Exception as e:
        debug(f'{vsim_path} -version: exception {e}')

    if proc is None or proc.returncode != 0:
        debug(f'{vsim_path} returncode non-zero: {proc=}')
        return

    stdout_str_lower = proc.stdout.decode('utf-8', errors='replace').lower()
    dirpath = os.path.abspath(dirpath)
    settool = None

    if all(x in stdout_str_lower for x in ('starter', 'modelsim', 'fpga')):
        settool = 'modelsim_ase'
    elif all(x in stdout_str_lower for x in ('starter', 'questa', 'fpga')):
        settool = 'questa_fse'
    elif all(x in stdout_str_lower for x in ('questa', 'fpga')):
        settool = 'questa_fe'
    elif all(x in stdout_str_lower for x in ('riviera', 'aldec')):
        settool = 'riviera'
    elif 'questa' in stdout_str_lower:
        settool = 'questa'

    debug(f'{vsim_path=} {settool=} from version')
    if settool not in TOOL_PATH:
        TOOL_PATH[settool] = None

    # TOOL_PATH[settool] default is None, so if found, set it
    # if TOOL_PATH[settool] is already set and not None, don't set again.
    if settool and not TOOL_PATH.get(settool, None):
        TOOL_PATH[settool] = vsim_path

        version = ''
        for word in stdout_str_lower.split():
            if all(c.isdigit() or c == '.' for c in word):
                parts = word.split('.')
                if len(parts) >= 2:
                    version = word
                    break

        if version:
            parts = version.split('.')
            if len(parts) > 3:
                # stop after 3 numbers
                version = '.'.join(parts[:3])
            TOOL_VERSION[settool] = version


def init() -> None:
    '''Sets INIT_HAS_RUN=True (only runs once) and updates TOOL_PATH[tool]'''
    global INIT_HAS_RUN # pylint: disable=global-statement

    if INIT_HAS_RUN:
        return

    INIT_HAS_RUN = True
    vsim_path = safe_shutil_which('vsim') # It might be vsim.EXE in Windows.
    debug(f'vsim_helper: found vsim executable {vsim_path}')

    if not vsim_path:
        return

    path_env = os.environ.get('PATH', '')
    paths = path_env.split(':')

    for path in paths:
        if not os.path.isdir(path):
            continue
        files = os.listdir(path)
        if any(x in files for x in ('vsim', 'vsim.exe', 'vsim.bat')):
            debug(f'vsim_helper: found in PATH: vsim executable {vsim_path} in {path}')
            init_get_version(dirpath=path)

    debug(f'vsim_helper: {TOOL_PATH=}')



''' opencos.tools.cocotb - Used by opencos.eda for sim commands with --tool=cocotb.

Contains classes for ToolCocotb, CommandSimCocotb.
'''

import os
from importlib import metadata

from opencos import util
from opencos.commands import CommandSim
from opencos.eda_base import Tool
from opencos.files import safe_shutil_which, PY_EXE
from opencos.utils import status_constants
from opencos.utils.str_helpers import sanitize_defines_for_sh
from opencos.tools import verilator # For default waivers.


class ToolCocotb(Tool):
    '''ToolCocotb used by opencos.eda for --tool=cocotb'''

    _TOOL = 'cocotb'
    _EXE = 'python'
    _URL = 'https://github.com/cocotb/cocotb'

    cocotb_version = ''
    python_exe = ''

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        # Check if python is available
        python_path = PY_EXE
        if not python_path:
            self.error('"python" or "python3" not in path, required for cocotb')
        else:
            self.python_exe = python_path

        # Check if cocotb is installed
        try:
            version = metadata.version('cocotb')
            self.cocotb_version = version
            self._VERSION = version
            util.debug(f'Found cocotb version: {version}')
            return self._VERSION
        except metadata.PackageNotFoundError:
            util.warning('cocotb package not installed in python environment. '
                       'Install with: pip install cocotb; or pypackage: opencos-eda[cocotb]')
        except Exception as e:
            util.warning(f'Failed to check cocotb version: {e}')

        return ''


class CommandSimCocotb(CommandSim, ToolCocotb):
    '''CommandSimCocotb is a command handler for: eda sim --tool=cocotb'''

    def __init__(self, config: dict):
        CommandSim.__init__(self, config)
        ToolCocotb.__init__(self, config=self.config)

        self.args.update({
            'gui': False,
            'tcl-file': None,
            'cocotb-test-module': None,
            'cocotb-test-runner': 'python',
            'cocotb-test-runner-file': '',
            'cocotb-test-run-dir': None, # If None, use self.args['work-dir']
            'cocotb-simulator': 'verilator',
            'cocotb-makefile': False,
            'cocotb-python-runner': True,
            'cocotb-standalone-makefile': False,
        })

        self.args_help.update({
            'waves': 'Include waveforms by setting COCOTB_ENABLE_WAVES=1',
            'cocotb-test-module': 'Python test module name (e.g., test_my_design)',
            'cocotb-test-runner': 'Test runner to use: python (default) or pytest',
            'cocotb-test-runner-file': 'Bring your own test_runner.py file',
            'cocotb-test-run-dir': (
                'Directory where cocotb-test-runner will run, if unset uses default eda.work/(name)'
            ),
            'cocotb-simulator': (
                'Simulator backend: verilator (default), icarus, etc.'
                ' Note that iverilog will convert to icarus here'
            ),
            'cocotb-makefile': 'Use traditional Makefile system instead of Python runner',
            'cocotb-python-runner': 'Use Python-based runner system (default, cocotb 1.8+)',
            'cocotb-standalone-makefile': (
                'Use provided Makefile as-is, run make in source directory'
            ),
        })

        self.cocotb_command_lists = []
        self.cocotb_test_files = []


    def help( # pylint: disable=dangerous-default-value
            self, tokens: list = [], no_targets: bool = False,
            show_all_eda_commands: bool = False
    ) -> None:
        '''Override for Command.help(...)'''

        super().help(tokens=tokens, no_targets=no_targets)
        self._warn_if_simulator_not_present()

    def _warn_if_simulator_not_present(self) -> None:
        '''Warn if --cocotb-simulator is not set, or if the exe is not in PATH'''

        simulator = self.args['cocotb-simulator']
        if not simulator:
            util.warning('--cocotb-simulator is not set, a simulation cannot be run with'
                         'this arg value')
            return
        exe = safe_shutil_which(simulator)
        if not exe:
            util.warning(f'--cocotb-simulator={simulator}, {simulator} is not present in PATH',
                         'a simulation cannot be run with this arg value')


    def prepare_compile(self) -> None:
        self.set_tool_defines()

        # Check existence of cocotb-simulator:
        self._warn_if_simulator_not_present()

        # Fix iverilog -> icarus
        if self.args['cocotb-simulator'] == 'iverilog':
            self.args['cocotb-simulator'] = 'icarus'

        # Find cocotb test files
        self._find_cocotb_test_files()

        if self.args['cocotb-standalone-makefile']:
            self._prepare_standalone_makefile_system()
        elif self.args['cocotb-makefile']:
            self._prepare_makefile_system()
        else:
            self._prepare_python_runner_system()

        # Create directories
        paths = ['logs', 'sim_build']
        util.safe_mkdirs(base=self.args['work-dir'], new_dirs=paths)

        # Write shell scripts
        self.write_sh_scripts_to_work_dir(
            compile_lists=[],
            elaborate_lists=[],
            simulate_lists=self.cocotb_command_lists,
            simulate_line_breaks=True,
            simulate_sh_fname='cocotb_test.sh'
        )


    def _find_cocotb_test_files(self):
        '''Find Python test files that contain cocotb tests'''
        self.cocotb_test_files = []

        # Look for test files in the current directory and deps
        for file_path in self.files_non_source + self.files_py:
            if (file_path.endswith('.py') and
                ('test' in file_path.lower() or 'tb' in file_path.lower())):
                # Check if it's a cocotb test file
                try:
                    with open(file_path, 'r', encoding='utf-8') as f:
                        content = f.read()
                        if 'import cocotb' in content or 'from cocotb' in content:
                            self.cocotb_test_files.append(file_path)
                            util.debug(f'Found cocotb test file: {file_path}')
                except Exception as e:
                    util.debug(f'Could not read {file_path}: {e}')

        # If no test files found, look for explicit test module
        if not self.cocotb_test_files and self.args.get('cocotb-test-module'):
            test_module = self.args['cocotb-test-module']
            if not test_module.endswith('.py'):
                test_module += '.py'

            # Look in work directory and target directory
            for search_dir in [self.args['work-dir'], os.path.dirname(self.target or '.')]:
                test_path = os.path.join(search_dir, test_module)
                if os.path.exists(test_path):
                    self.cocotb_test_files.append(test_path)
                    break

        if not self.cocotb_test_files:
            self.error('No cocotb test files found. Either include .py files with '
                      'cocotb imports in your deps, or specify --cocotb-test-module')

    def _prepare_python_runner_system(self):
        '''Prepare cocotb using the Python-based runner system (cocotb 1.8+)'''

        # Create a Python runner script, or use from --cocotb-test-runner-file
        if self.args['cocotb-test-runner-file']:

            runner_script = self.args['cocotb-test-runner-file']
            if not os.path.isfile(self.args['cocotb-test-runner-file']):
                self.error(
                    "File does not exist, for: --cocotb-test-runner-file=",
                    f"{self.args['cocotb-test-runner-file']}",
                    error_code=status_constants.EDA_GENERAL_FILE_NOT_FOUND
                )
                return
        else:
            runner_script = self._create_python_runner_script()

        if self.args['cocotb-test-runner'] == 'pytest':
            # Use pytest to run the tests
            runner_script_name = os.path.basename(runner_script)
            cmd_list = [
                self.python_exe, '-m', 'pytest',
                runner_script_name,
                '-v', '-s'
            ]
        else:
            # Run the Python script directly
            runner_script_name = os.path.basename(runner_script)
            cmd_list = [self.python_exe, runner_script_name]

        # Set environment variables
        env_vars = self._get_cocotb_env_vars()

        # Create command with environment variables, needs to be list-of-(cmd_list)
        self.cocotb_command_lists = [
            util.ShellCommandList(
                self._create_env_command(env_vars) + cmd_list,
                tee_fpath='cocotb_test.log',
                # If someone wanted to run this from a specific dir, then
                # run it from there.
                work_dir=self.args['cocotb-test-run-dir']
            )
        ]

    def _prepare_makefile_system(self):
        '''Prepare cocotb using the traditional Makefile system'''

        makefile_path = os.path.join(self.args['work-dir'], 'Makefile')
        self.files_makefile.append(makefile_path)
        with open(makefile_path, 'w', encoding='utf-8') as f:
            f.write(self._create_makefile_content())

        cmd_list = self._create_shell_command_with_success('make -f Makefile')
        self.cocotb_command_lists = [
            util.ShellCommandList(
                cmd_list, tee_fpath='cocotb_makefile.log',
                # If someone wanted to run this from a specific dir, then
                # run it from there.
                work_dir=self.args['cocotb-test-run-dir']
            )
        ]

    def _prepare_standalone_makefile_system(self):
        '''Use provided Makefile as-is, run make in source directory'''

        # Find the Makefile in our dependencies
        makefile_path = None
        for file_path in self.files_non_source + self.files_makefile:
            if os.path.basename(file_path).lower() == 'makefile':
                makefile_path = file_path
                break

        if not makefile_path:
            self.error('No Makefile found in deps for --cocotb-standalone-makefile')

        if any(value for key,value in self.args.items() if key.startswith('export')):
            util.warning(f'Using an --export arg with Cocotb standalong Makefile {makefile_path}',
                         'is not fully supported')

        makefile_dir = os.path.dirname(os.path.abspath(makefile_path))
        cmd_list = self._create_shell_command_with_success('make')
        self.cocotb_command_lists = [
            util.ShellCommandList(
                cmd_list,
                tee_fpath='cocotb_standalone.log',
                work_dir=makefile_dir
            )
        ]

    def _get_test_module_name(self) -> str:
        '''Get the test module name from args or detected files'''
        if self.args.get('cocotb-test-module'):
            return self.args['cocotb-test-module']
        if self.cocotb_test_files:
            # Use the first test file found, strip .py extension
            test_file = os.path.basename(self.cocotb_test_files[0])
            return test_file.replace('.py', '')
        return 'test_design'


    def _get_hdl_sources(self) -> dict:
        '''Get HDL source files organized by type'''
        return {
            'verilog': (self.files_sv or []) + (self.files_v or []),
            'vhdl': self.files_vhd or [],
            'all': (self.files_sv or []) + (self.files_v or []) + (self.files_vhd or [])
        }

    def _create_python_runner_script(self) -> str:
        '''Create the Python runner script and return its path'''
        test_module = self._get_test_module_name()
        hdl_sources = self._get_hdl_sources()['all']

        script_content = self._generate_runner_script_content(test_module, hdl_sources)
        script_path = os.path.join(self.args['work-dir'], 'cocotb_runner.py')

        with open(script_path, 'w', encoding='utf-8') as f:
            f.write(script_content)

        os.chmod(script_path, 0o755)

        return script_path


    def _generate_runner_script_content(self, test_module: str, hdl_sources: list) -> str:
        '''Generate the content for the Python runner script'''

        if safe_shutil_which('verilator'):
            tmp_verilator_obj = verilator.VerilatorSim(config=self.config)
            verilator_waivers = tmp_verilator_obj.get_verilator_tool_config_waivers()
        else:
            verilator_waivers = []

        return f'''#!/usr/bin/env python3
"""
Cocotb test runner script generated by opencos
"""

import os
import sys
from pathlib import Path

# Add current directory to Python path for test imports
sys.path.insert(0, os.getcwd())

try:
    from cocotb_tools.runner import get_runner
except ImportError:
    print("ERROR: cocotb not found. Install with: pip install cocotb")
    sys.exit(1)

def run_cocotb_test():
    """Run cocotb test using the Python runner system"""

    # Configuration
    simulator = "{self.args['cocotb-simulator']}"
    hdl_toplevel = "{self.args.get('top', 'top')}"
    test_module = "{test_module}"

    # HDL source files
    hdl_sources = {hdl_sources!r}

    # Convert to Path objects and make absolute
    sources = []
    for src in hdl_sources:
        src_path = Path(src)
        if not src_path.is_absolute():
            src_path = Path.cwd() / src_path
        sources.append(src_path)

    # Include directories
    include_dirs = {list(self.incdirs)!r}

    # Defines (filter out None values for cocotb compatibility)
    all_defines = {dict(self.defines)!r}
    defines = {{k: v for k, v in all_defines.items() if v is not None}}

    # Parameters (empty for simple modules without parameters)
    parameters = {{}}

    try:
        # Get the runner for the specified simulator
        runner = get_runner(simulator)

        build_args = []

        if simulator == "verilator":
            build_args.extend({list(self.args.get('verilate-args', []))!r})
            build_args.extend({list(self.args.get('compile-args', []))!r})
            build_args.extend({list(verilator_waivers)!r})

        # Build the design
        runner.build(
            sources=sources,
            hdl_toplevel=hdl_toplevel,
            includes=include_dirs,
            defines=defines,
            parameters=parameters,
            build_dir="sim_build",
            build_args=build_args,
        )

        # Run the test
        runner.test(
            hdl_toplevel=hdl_toplevel,
            test_module=test_module,
            test_dir=".",
        )

        print("{self._get_success_message()}")

    except Exception as e:
        print(f"ERROR: Cocotb test failed: {{e}}")
        sys.exit(1)

if __name__ == "__main__":
    run_cocotb_test()
'''

    def _create_makefile_content(self) -> str:
        '''Create Makefile content for cocotb traditional system'''

        # Determine test module
        test_module = self._get_test_module_name()

        # Get HDL sources and determine language
        hdl_sources = self._get_hdl_sources()
        verilog_sources = hdl_sources['verilog']
        vhdl_sources = hdl_sources['vhdl']

        # Determine HDL language
        if vhdl_sources and not verilog_sources:
            hdl_lang = 'vhdl'
            sources_var = 'VHDL_SOURCES'
            sources_list = ' '.join(vhdl_sources)
        else:
            hdl_lang = 'verilog'
            sources_var = 'VERILOG_SOURCES'
            sources_list = ' '.join(verilog_sources)

        makefile_content = f'''# Cocotb Makefile generated by opencos

# Simulator selection
SIM ?= {self.args['cocotb-simulator']}

# HDL language
TOPLEVEL_LANG = {hdl_lang}

# HDL sources
{sources_var} = {sources_list}

# Top level module
TOPLEVEL = {self.args.get('top', 'top')}

# Test module
MODULE = {test_module}

# Include directories
'''

        if self.incdirs:
            makefile_content += ('COMPILE_ARGS += ' +
                               ' '.join(f'-I{inc}' for inc in self.incdirs) + '\n')

        # Add defines
        if self.defines:
            define_args = []
            for k, v in self.defines.items():
                if v is None:
                    define_args.append(f'-D{k}')
                else:
                    define_args.append(f'-D{k}={sanitize_defines_for_sh(v)}')
            makefile_content += 'COMPILE_ARGS += ' + ' '.join(define_args) + '\n'

        if self.args['cocotb-simulator'] == 'verilator' and self.args.get('verilate-args'):
            makefile_content += 'COMPILE_ARGS += ' + ' '.join(self.args['verilate-args']) + '\n'

        makefile_content += '''
# Waves support
ifeq ($(WAVES),1)
    COCOTB_ENABLE_WAVES = 1
    export COCOTB_ENABLE_WAVES
endif

# Include cocotb's Makefile
include $(shell cocotb-config --makefiles)/Makefile.sim
'''

        return makefile_content

    def _get_success_message(self) -> str:
        '''Get standardized success message'''
        return "Cocotb test completed successfully!"

    def _create_shell_command_with_success(self, base_command: str) -> list:
        '''Create a shell command list with success message'''
        return ['sh', '-c', f'{base_command} && echo "{self._get_success_message()}"']

    def _get_cocotb_env_vars(self) -> dict:
        '''Get environment variables for cocotb execution'''
        env_vars = {}

        top = self.args.get('top', 'top')

        # Basic cocotb configuration
        env_vars.update({
            'SIM': self.args['cocotb-simulator'],
            'TOPLEVEL': top,
            'COCOTB_TOPLEVEL': top,
        })

        if test_module := self.args['cocotb-test-module']:
            env_vars.update({
                'MODULE': test_module,
                'COCOTB_TEST_MODULES': test_module,
            })

        # Enable waves if requested
        if self.args.get('waves', False):
            env_vars['COCOTB_ENABLE_WAVES'] = '1'

        # Set log level based on verbosity
        if util.args.get('verbose', False):
            env_vars['COCOTB_LOG_LEVEL'] = 'DEBUG'
        else:
            env_vars['COCOTB_LOG_LEVEL'] = 'INFO'

        # Random seed
        if seed := self.args.get('seed'):
            env_vars['COCOTB_RANDOM_SEED'] = str(seed)

        return env_vars

    def _create_env_command(self, env_vars: dict) -> list:
        '''Create environment variable command prefix'''
        env_cmd = []
        for key, value in env_vars.items():
            env_cmd.extend(['env', f'{key}={value}'])
        return env_cmd

    def compile(self):
        # For cocotb, compilation happens as part of the test run
        if self.args['stop-before-compile']:
            return
        util.info('Cocotb: compilation will happen during test execution')

    def elaborate(self):
        # For cocotb, elaboration happens as part of the test run
        pass

    def simulate(self):
        if self.args['stop-before-compile'] or self.args['stop-after-compile'] or \
           self.args['stop-after-elaborate']:
            return

        # Run the cocotb tests
        self.run_commands_check_logs(self.cocotb_command_lists)

    def get_compile_command_lists(self, **kwargs) -> list:
        # Cocotb handles compilation internally
        return []

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # Cocotb handles elaboration internally
        return []

    def get_simulate_command_lists(self, **kwargs) -> list:
        return self.cocotb_command_lists

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []



'''opencos.commands.shell - Base class command handler for: eda shell ...

Not intended to be overriden by Tool based classes.'''

# Note - tricky to eliminate it with inheritance when calling reusable methods.
# pylint: disable=R0801

# TODO(drew): clean up CommandShell.check_logs_for_errors and CommandShell.run_commands_check_logs
#   These also share a lot with CommandSim.* methods, so consider refactoring to share code,
#   for example, CommandShell.do_export could move to CommandDesign, and derived classes
#   set the allow-listed args to pass to export.
# pylint: disable=too-many-arguments

import os

from opencos import util
from opencos.eda_base import CommandDesign

from opencos.utils import status_constants


class CommandShell(CommandDesign):
    '''Base class command handler for: eda shell ...'''

    command_name = 'shell'

    def __init__(self, config: dict):
        CommandDesign.__init__(self, config=config)
        self.args.update({
            'pass-pattern': "",
            'log-bad-strings': [],
            'log-must-strings': [],
        })
        self.args_help.update({
            'log-bad-strings': (
                'strings that if present in the log will cause an `eda shell` error'
            ),
            'log-must-strings': (
                'strings that are required by the log to not-fail the `eda shell` call'
            ),
            'pass-pattern': (
                'Additional string required to pass the `eda shell` call, appends to'
                ' log-must-strings'
            ),
        })

    def process_tokens(self, tokens: list, process_all: bool = True,
                       pwd: str = os.getcwd()) -> list:
        unparsed = CommandDesign.process_tokens(
            self, tokens=tokens, process_all=process_all, pwd=pwd
        )

        if self.stop_process_tokens_before_do_it():
            return unparsed

        if self.args['top']:
            # create our work dir
            self.create_work_dir()
            self.run_dep_commands()
            self.do_it()
            self.run_post_tool_dep_commands()
        else:
            util.warning(f'CommandShell: {self.command_name=} not run due to lack of',
                         f'{self.args["top"]=} value')
        return unparsed


    def run_commands_check_logs( # pylint: disable=dangerous-default-value
            self, commands: list , check_logs: bool = True, log_filename=None,
            bad_strings: list = [],
            must_strings: list = [],
            use_bad_strings: bool = True, use_must_strings: bool = True
    ) -> None:
        '''Returns None, runs all commands (each element is a list) and checks logs

        for bad-strings and must-strings (args or class member vars)
        '''

        for obj in commands:

            assert isinstance(obj, list), \
                (f'{self.target=} command {obj=} is not a list or util.ShellCommandList,'
                 ' not going to run it.')

            clist = list(obj).copy()
            tee_fpath = getattr(obj, 'tee_fpath', None)

            util.debug(f'run_commands_check_logs: {clist=}, {tee_fpath=}')

            log_fname = None
            if tee_fpath:
                log_fname = tee_fpath
            if log_filename:
                log_fname = log_filename

            self.exec(work_dir=self.args['work-dir'], command_list=clist, tee_fpath=tee_fpath)

            if check_logs and log_fname:
                self.check_logs_for_errors(
                    filename=log_fname, bad_strings=bad_strings, must_strings=must_strings,
                    use_bad_strings=use_bad_strings, use_must_strings=use_must_strings
                )


    def do_export(self, **kwargs) -> None:
        '''CommandShell helper for handling args --export*, override from Command'''
        super().do_export(
            deps_file_args=[
                '--log-must',
                '--log-bad',
                '--pass-pattern'
            ],
            support_export_run=True
        )


    def do_it(self) -> None:
        self.write_eda_config_and_args()

        if self.is_export_enabled():
            # If we're exporting the target, we do NOT run the test here
            # (do_export() may run the test in a separate process and
            # from the out_dir if --export-run was set)
            self.do_export()


    def check_logs_for_errors( # pylint: disable=dangerous-default-value
            self, filename: str,
            bad_strings: list = [], must_strings: list = [],
            use_bad_strings: bool = True, use_must_strings: bool = True
    ) -> None:
        '''Returns None, checks logs using args bad_strings, must_strings,

        and internals self.args["log-[bad|must]-strings"] (lists).
        '''

        _bad_strings = bad_strings
        _must_strings = must_strings
        # append, if not they would 'replace' the args values:
        if use_bad_strings:
            _bad_strings = bad_strings + self.args.get('log-bad-strings', [])
        if use_must_strings:
            _must_strings = must_strings + self.args.get('log-must-strings', [])

        if self.args['pass-pattern'] != "":
            _must_strings.append(self.args['pass-pattern'])

        if len(_bad_strings) > 0 or len(_must_strings) > 0:
            hit_must_string_dict = dict.fromkeys(_must_strings)
            fname = os.path.join(self.args['work-dir'], filename)
            with open(fname, 'r', encoding='utf-8') as f:
                for lineno, line in enumerate(f):
                    if any(must_str in line for must_str in _must_strings):
                        for k, _ in hit_must_string_dict.items():
                            if k in line:
                                hit_must_string_dict[k] = True
                    if any(bad_str in line for bad_str in _bad_strings):
                        self.error(
                            f"log {fname}:{lineno} contains one of {_bad_strings=}",
                            error_code=status_constants.EDA_SHELL_LOG_HAS_BAD_STRING
                        )

            if any(x is None for x in hit_must_string_dict.values()):
                self.error(
                    f"Didn't get all passing patterns in log {fname}: {_must_strings=}",
                    f" {hit_must_string_dict=}",
                    error_code=status_constants.EDA_SHELL_LOG_MISSING_MUST_STRING
                )



''' opencos.utils.markup_helpers - function helpers for YAML, TOML reading and writing'''

import json
import os
import re
import subprocess

import yaml

from opencos.files import safe_shutil_which
from opencos.util import debug, error, info, warning


def yaml_load_only_root_line_numbers(filepath: str) -> dict:
    '''Returns a dict of {key: int line number}, very crude'''
    # Other solutions aren't as attractive, require a lot of mappers to get
    # line numbers on returned values that aren't dict
    data = None
    with open(filepath, encoding='utf-8') as f:
        try:
            # Try to do a very lazy parse of root level keys only, returns dict{key:lineno}
            data = {}
            for lineno,line in enumerate(f.readlines()):
                m = re.match(r'^(\w+):', line)
                if m:
                    key = m.group(1)
                    data[key] = lineno + 1
        except Exception as e:
            error(f"Error loading YAML {filepath=}:", e)
    return data


def toml_load_only_root_line_numbers(filepath: str) -> dict:
    '''Returns a dict of {key: int line number}, very crude'''
    data = None
    with open(filepath, encoding='utf-8') as f:
        try:
            data = {}
            for lineno, line in enumerate(f.readlines()):
                m = re.match(r'^\[(\w+)\]', line)
                if m:
                    key = m.group(1)
                    data[key] = lineno + 1
        except Exception as e:
            error(f'Error loading TOML {filepath=}', e)
    return data


def yaml_safe_load(filepath: str, only_root_line_numbers:bool = False) -> dict:
    '''Returns dict or None from filepath (str), errors if return type not in assert_return_types.

    only_root_line_numbers -- if True, will return a dict of {key: line number (int)} for
                              all the root level keys. Used for debugging DEPS.yml in
                              eda.CommandDesign.resolve_target_core
    '''

    data = None

    if only_root_line_numbers:
        return yaml_load_only_root_line_numbers(filepath)

    with open(filepath, encoding='utf-8') as f:
        debug(f'Opening {filepath=}')
        try:
            data = yaml.safe_load(f)
        except yaml.YAMLError as e:

            # if yamllint is installed, then use it to get all errors in the .yml|.yaml
            # file, instead of the single exception.
            if safe_shutil_which('yamllint'):
                try:
                    sp_out = subprocess.run(
                        f'yamllint -d relaxed --no-warnings {filepath}'.split(),
                        capture_output=True, text=True, check=False )
                    for x in sp_out.stdout.split('\n'):
                        if x:
                            info('yamllint: ' + x)
                except Exception as e2:
                    debug(f'yamllimt exception: {e2}')

            if hasattr(e, 'problem_mark'):
                mark = e.problem_mark
                error(f"Error parsing {filepath=}: line {mark.line + 1},",
                      f"column {mark.column +1}: {e.problem}")
            else:
                error(f"Error loading YAML {filepath=}:", e)
        except Exception as e:
            error(f"Error loading YAML {filepath=}:", e)

    return data


def yaml_safe_writer(data: dict, filepath: str) -> None:
    '''Wrapper for yaml.dump, enforces file extension otherwise warning'''
    _, ext = os.path.splitext(filepath)
    if ext.lower() in ('.yml', '.yaml'):
        try:
            with open(filepath, 'w', encoding='utf-8') as f:
                yaml.dump(data, f, allow_unicode=True,
                          default_flow_style=False, sort_keys=False, encoding='utf-8')
        except PermissionError:
            error(f'PermissionError trying to write file: {filepath}')
        except IOError as e:
            error(f'IOError while trying to write file: {filepath}, {e}')

    else:
        warning(f'{filepath=} to be written for this extension not implemented.')

def json_writer(data: dict, filepath: str) -> None:
    '''Wrapper for json.dump'''
    try:
        with open(filepath, 'w', encoding='utf-8') as f:
            json.dump(data, f, indent=4)
    except PermissionError:
        error(f'PermissionError trying to write file: {filepath}')
    except IOError as e:
        error(f'IOError while trying to write file: {filepath}, {e}')


def markup_writer(data: dict, filepath: str) -> None:
    '''Wrapper for yaml_safe_writer or json_writer'''
    _, ext = os.path.splitext(filepath)
    if ext.lower() in ('.yml', '.yaml'):
        yaml_safe_writer(data, filepath)
    else:
        json_writer(data, filepath)


def markup_dumper(data: dict, as_yaml: bool = False) -> str:
    '''Returns JSON str; if as_yaml=True returns YAML str, from data'''
    if as_yaml:
        return yaml.dump(
            data=data, allow_unicode=True,
            default_flow_style=False, sort_keys=False
        )

    # else return JSON:
    return str(json.dumps(data, indent=4))
